from __future__ import annotations

import io
from abc import ABC, abstractmethod
try:
    from functools import cache
except ImportError:
    # `functools.cache` does not exists on Python3.8
    from functools import lru_cache
    cache = lru_cache(maxsize=None)
from PIL import Image
from socket import socket
from threading import Lock
from typing import Any, Dict, IO, List, Optional, Tuple, Union

from speculos.observer import TextEvent
from .struct import DisplayArgs, MODELS, Pixel, ServerArgs

PixelColorMapping = Dict[Pixel, int]


class IODevice(ABC):
    """
    An interface every class implementing application IOs (screens, buttons, APDUs, ...) should
    inherit from.

    These classes will be managed by a `DisplayNotifier`, which is responsible to redirect events
    to the correct `IODevice` instance through its `IODevice.can_read` callback.
    """

    @property
    @abstractmethod
    def file(self) -> Union[IO[bytes], socket]:
        """
        Returns the file (Pipe, Socket, ...) tied to this `IODevice`
        """
        raise NotImplementedError()

    @property
    def fileno(self) -> int:
        """
        Returns the file descriptor of the file tied to this `IODevice`
        """
        return self.file.fileno()

    @abstractmethod
    def can_read(self, screen: DisplayNotifier) -> None:
        """
        Callback used by a notifier to trigger `IODevice` events on given screen
        """
        pass


COLORS: Dict[str, int] = {
    'LAGOON_BLUE': 0x7ebab5,
    'JADE_GREEN': 0xb9ceac,
    'FLAMINGO_PINK': 0xd8a0a6,
    'SAFFRON_YELLOW': 0xf6a950,
    'MATTE_BLACK': 0x111111,
    'CHARLOTTE_PINK': 0xff5555,
    'ARNAUD_GREEN': 0x79ff79,
    'SYLVE_CYAN': 0x29f3f3,
}


class FrameBuffer:
    """
    A class responsible for managing the graphic screen of the current application.

    It updates the screen, takes screenshots, manages colors and such.
    """

    COLORS = {
        "nanos": 0x00fffb,
        "nanox": 0xdddddd,
        "nanosp": 0xdddddd,
        "stax": 0xdddddd,
        "flex": 0xdddddd,
    }

    def __init__(self, model: str):
        self.pixels: PixelColorMapping = {}
        self.screenshot_pixels: PixelColorMapping = {}
        self.screenshot_pixels_lock = Lock()
        self.default_color = 0
        self.draw_default_color = False
        self.reset_screeshot_pixels = False
        self._public_screenshot_value = b''
        self.current_data = b''
        self.recreate_public_screenshot = True
        self.model = model
        self.current_screen_size = MODELS[model].screen_size
        self._width, self._height = MODELS[model].screen_size

    @cache
    def check_color(self, color: int) -> int:
        # There are only 2 colors on the Nano S and the Nano X but the one
        # passed in argument isn't always valid. Fix it here.
        if self.model != 'stax' and self.model != 'flex':
            if color != 0x000000:
                color = FrameBuffer.COLORS.get(self.model, color)
        return color

    def draw_point(self, x: int, y: int, color: int) -> None:
        self.pixels[(x, y)] = self.check_color(color)

    def draw_horizontal_line(self, x0: int, y: int, width: int, color: int) -> None:
        for x in range(x0, x0 + width):
            self.pixels[(x, y)] = self.check_color(color)

    def draw_rect(self, x0: int, y0: int, width: int, height: int, color: int) -> List[TextEvent]:
        color = self.check_color(color)

        if x0 == 0 and y0 == 0 and width == self._width and height == self._height:
            self.default_color = color
            self.draw_default_color = True
            self.pixels = {}
            self.reset_screeshot_pixels = True
            return [TextEvent("", 0, 0, 0, 0, True)]

        for x in range(x0, x0 + width):
            for y in range(y0, y0 + height):
                self.pixels[(x, y)] = color

        return []

    def _get_image(self) -> bytes:
        # This call is made from the Speculos API thread
        # Protect screenshot_pixels for concurrent Write during this Read
        with self.screenshot_pixels_lock:
            data = bytearray(self.default_color.to_bytes(3, "big")) * self._width * self._height
            for (x, y), color in self.screenshot_pixels.items():
                pos = 3 * (y * self._width + x)
                data[pos:pos + 3] = color.to_bytes(3, "big")
        return bytes(data)

    def _get_screenshot_iobytes_value(self) -> bytes:
        # Get the pixels object once, as it may be replaced during the loop.
        data = self._get_image()

        image = Image.frombytes("RGB", self.current_screen_size, data)
        iobytes = io.BytesIO()
        image.save(iobytes, format="PNG")
        return iobytes.getvalue()

    def take_screenshot(self) -> Tuple[Tuple[int, int], bytes]:
        return self.current_screen_size, self._get_image()

    def update_screenshot(self) -> None:
        # This call is made from the MCU/Seproxyhal thread
        # Protect screenshot_pixels for concurrent Read during this Write
        with self.screenshot_pixels_lock:
            if self.reset_screeshot_pixels:
                self.screenshot_pixels = {}
                self.reset_screeshot_pixels = False
            self.screenshot_pixels.update(self.pixels)

    def update_public_screenshot(self) -> None:
        # Stax/Flex only
        # As we lazyly evaluate the published screenshot, we only flag the evaluation update as necessary
        self.recreate_public_screenshot = True

    @property
    def public_screenshot_value(self) -> bytes:
        # Stax/Flex only
        # Lazy calculation of the published screenshot, as it is a costly operation
        # and not necessary if no one tries to read the value
        if self.recreate_public_screenshot:
            self.recreate_public_screenshot = False
            self._public_screenshot_value = self._get_screenshot_iobytes_value()

        return self._public_screenshot_value

    def get_public_screenshot(self) -> bytes:
        if self.model == "stax" or self.model == "flex":
            # On Stax/Flex, we only make the screenshot public on the RESTFUL api when it is consistent with events
            # On top of this, take_screenshot is time consuming on stax/flex, so we'll do as few as possible
            # We return the value calculated last time update_public_screenshot was called
            return self.public_screenshot_value
        # On nano we have no knowledge of screen refreshes so we can't be scarce on publishes
        # So we publish the raw current content every time. It's ok as take_screenshot is fast on Nano
        return self._get_screenshot_iobytes_value()

    # Should be declared as an `@abstractmethod` (and also `class FrameBuffer(ABC):`), but in this
    # case multiple inheritance in `screen.PaintWidget(FrameBuffer, QWidget)` will break, as both
    # FrameBuffer and QWidget derive from a different metaclass, and Python cannot figure out which
    # class builder to use.
    def update(self,
               x: Optional[int] = None,
               y: Optional[int] = None,
               w: Optional[int] = None,
               h: Optional[int] = None) -> bool:
        raise NotImplementedError()


class GraphicLibrary(ABC):
    """
    A class interface defining mandatory method a graphical library must implement.

    Currently implemented graphic libraries are `bagl.Bagl` and `nbgl.NBGL`.
    """

    def __init__(self, fb: FrameBuffer, size: Tuple[int, int], model: str):
        self._fb = fb
        self.SCREEN_WIDTH, self.SCREEN_HEIGHT = size
        self.model = model

    @property
    def fb(self) -> FrameBuffer:
        return self._fb

    @abstractmethod
    def refresh(self, data: bytes) -> bool:
        pass

    def update_screenshot(self) -> None:
        self.fb.update_screenshot()

    def update_public_screenshot(self) -> None:
        self.fb.update_public_screenshot()

    def take_screenshot(self) -> Tuple[Tuple[int, int], bytes]:
        return self.fb.take_screenshot()


class Display(ABC):
    """
    A class interface for managing the graphic display of an application.

    Every display type is composed of two classes:
    - A `Display` implementation, which will mostly deal with the graphics
    - A `DisplayNotifier`, which will managed the `IOdevice` tied to the running application.

    Currently, there are 3 display management type, stored in following modules:
    - `screen.py`, displaying the application infos using Qt
    - `screen_text.py`, displaying the application infos using the terminal (through `curses`)
    - `headless.py`, displaying nothing, although the application interface can still be reached
      through VNC if activated.
    """

    def __init__(self, display_args: DisplayArgs, server_args: ServerArgs) -> None:
        self._server_args = server_args
        self._display_args = display_args

    @property
    def apdu(self) -> Any:  # ApduServer:
        return self._server_args.apdu

    @property
    def seph(self) -> Any:  # SeProxyHal:
        return self._server_args.seph

    @property
    def model(self) -> str:
        return self._display_args.model

    @property
    def rendering(self):
        return self._display_args.rendering

    @property
    def use_bagl(self) -> bool:
        return self._display_args.use_bagl

    @property
    @abstractmethod
    def gl(self) -> GraphicLibrary:
        pass

    @abstractmethod
    def display_status(self, data: bytes) -> List[TextEvent]:
        raise NotImplementedError()

    @abstractmethod
    def display_raw_status(self, data: bytes):
        pass

    @abstractmethod
    def screen_update(self) -> bool:
        pass

    def forward_to_app(self, packet: bytes) -> None:
        self.seph.to_app(packet)

    def forward_to_apdu_client(self, packet: bytes) -> None:
        self.apdu.forward_to_client(packet)


class DisplayNotifier(ABC):
    """
    A class interface for managing the events between an application display and the `IODevice`
    subclasses.

    Every display type is composed of two classes:
    - A `Display` implementation, which will mostly deal with the graphics
    - A `DisplayNotifier`, which will managed the `IOdevice` tied to the running application.

    Currently, there are 3 display management type, stored in following modules:
    - `screen.py`, displaying the application infos using Qt
    - `screen_text.py`, displaying the application infos using the terminal (through `curses`)
    - `headless.py`, displaying nothing, although the application interface can still be reached
      through VNC if activated.
    """

    def __init__(self, display_args: DisplayArgs, server_args: ServerArgs) -> None:
        # TODO: this should be Dict[int, IODevice], but in QtScreen, it is
        #       a QSocketNotifier, which has a completely different interface
        #       and is not used in the same way in the mcu/screen.py module.
        self.notifiers: Dict[int, Any] = {}
        self._server_args = server_args
        self._display_args = display_args
        self._display: Display
        self.__init_notifiers()

    def _set_display_class(self, display_class: type):
        self._display = display_class(self._display_args, self._server_args)

    @property
    def display(self) -> Display:
        return self._display

    def add_notifier(self, device: IODevice):
        assert device.fileno not in self.notifiers
        self.notifiers[device.fileno] = device

    def remove_notifier(self, fd: int):
        self.notifiers.pop(fd)

    def __init_notifiers(self) -> None:
        for device in self._server_args._asdict().values():
            if device:
                self.add_notifier(device)

    @abstractmethod
    def run(self) -> None:
        pass



#!/usr/bin/env python3

"""
Record an emulated device screen and save the animation as a gif.
"""

import argparse
import io
import logging
import requests
from PIL import Image


def take_screenshot(session, api_url):
    response = session.get(f"{api_url}/screenshot")
    assert response.status_code == 200
    return response.content


def record_images(api_url):
    images = []

    logging.info("press CTRL-C to stop recording")

    try:
        with requests.Session() as session:
            with session.get(f"{api_url}/events?stream=true", stream=True) as stream:
                while True:
                    image = take_screenshot(session, api_url)
                    if len(images) == 0 or image != images[-1]:
                        images.append(image)
                    line = stream.raw.readline()
                    logging.debug(f"got event: {line}")
                    if not line:
                        break
    except KeyboardInterrupt:
        pass

    return images


def save_gif(outfile, images, duration=500):
    logging.info(f"saving images to {outfile}")
    images = [Image.open(io.BytesIO(image)) for image in images]
    images[0].save(outfile, save_all=True, append_images=images[1:], duration=duration, loop=0)


if __name__ == "__main__":
    parser = argparse.ArgumentParser(description="Record a screen as a GIF.")
    parser.add_argument("--api-url", default="http://127.0.0.1:5000", help="REST API URL")
    parser.add_argument("--outfile", default="/tmp/speculos.gif", help="Output file")
    args = parser.parse_args()

    logging.basicConfig(level=logging.INFO)

    images = record_images(args.api_url)
    save_gif(args.outfile, images)



import speculos.resources_importer as r


def test_api_resource_exists():
    assert r.get_resources_path("api", "").exists()


def test_mcu_resource_exists():
    assert r.get_resources_path("mcu", "").exists()


def test_load_json():
    schema = r.get_resource_schema_as_json("api", "finger.schema")
    assert isinstance(schema, dict)



from flask import Response

from .restful import ScreenResource


class Screenshot(ScreenResource):
    def get(self):
        iobytes_value = self.screen.display.m.get_public_screenshot()
        response = Response(iobytes_value, mimetype="image/png")
        response.headers.add("Cache-control", "no-cache,no-store")
        return response



import pytest

from enum import IntEnum


CLA = 0xE0


class Ins(IntEnum):
    GET_PUBLIC_KEY = 0x40
    GET_VERSION = 0xC4


@pytest.fixture(scope="module")
def client(client_btc_testnet):
    return client_btc_testnet


def test_btc_lib(client):
    client.apdu_exchange(CLA, Ins.GET_VERSION, b"")



from unittest import TestCase

from speculos.client import Api


class TestApi(TestCase):

    def setUp(self):
        self.api_url = 'some random url'
        self.api = Api(self.api_url)

    def test_close_stream_None_should_not_raise(self):
        self.assertIsNone(self.api.stream)
        self.api.close_stream()



from __future__ import annotations

from PyQt5.QtWidgets import QApplication, QWidget, QMainWindow
from PyQt5.QtGui import QPainter, QColor, QPixmap
from PyQt5.QtGui import QIcon, QKeyEvent, QMouseEvent
from PyQt5.QtCore import QEvent, Qt, QSocketNotifier, QSettings, QRect
from PyQt5.sip import voidptr
from typing import List, Optional

from speculos.observer import TextEvent
from . import bagl
from . import nbgl
from .display import COLORS, Display, DisplayNotifier, FrameBuffer, GraphicLibrary, IODevice
from .readerror import ReadError
from .struct import DisplayArgs, MODELS, ServerArgs
from .vnc import VNC

BUTTON_LEFT = 1
BUTTON_RIGHT = 2
DEFAULT_WINDOW_X = 10
DEFAULT_WINDOW_Y = 10


class PaintWidget(FrameBuffer, QWidget):
    def __init__(self, parent, model: str, pixel_size: int, vnc: Optional[VNC] = None):
        QWidget.__init__(self, parent)
        FrameBuffer.__init__(self, model)
        self.pixel_size = pixel_size
        self.mPixmap = QPixmap()
        self.vnc = vnc

    def paintEvent(self, event: QEvent):
        if self.pixels or self.draw_default_color:
            pixmap = QPixmap(self.size() / self.pixel_size)
            pixmap.fill(Qt.white)
            painter = QPainter(pixmap)
            painter.drawPixmap(0, 0, self.mPixmap)
            self._redraw(painter)
            self.mPixmap = pixmap
            self.pixels = {}
            self.draw_default_color = False

        qp = QPainter(self)
        copied_pixmap = self.mPixmap
        if self.pixel_size != 1:
            # Only call scaled if needed.
            copied_pixmap = self.mPixmap.scaled(
                self.mPixmap.width() * self.pixel_size,
                self.mPixmap.height() * self.pixel_size)
        qp.drawPixmap(0, 0, copied_pixmap)

    def update(self,  # type: ignore[override]
               x: Optional[int] = None,
               y: Optional[int] = None,
               w: Optional[int] = None,
               h: Optional[int] = None) -> bool:
        if x and y and w and h:
            QWidget.update(self, QRect(x, y, w, h))
        else:
            QWidget.update(self)
        return self.pixels != {}

    def _redraw(self, qp):
        if self.draw_default_color:
            qp.fillRect(0, 0, self._width, self._height, QColor.fromRgb(self.default_color))

        for (x, y), color in self.pixels.items():
            qp.setPen(QColor.fromRgb(color))
            qp.drawPoint(x, y)

        if self.vnc is not None:
            self.vnc.redraw(self.pixels, self.default_color)

        self.update_screenshot()


class App(QMainWindow):
    def __init__(self, qt_app: QApplication, display: DisplayArgs, server: ServerArgs) -> None:
        super().__init__()
        self.setWindowTitle('Ledger %s Emulator' % MODELS[display.model].name)

        self.seph = server.seph
        self._width, self._height = MODELS[display.model].screen_size
        self.pixel_size = display.pixel_size
        self.box_position_x, self.box_position_y = MODELS[display.model].box_position
        box_size_x, box_size_y = MODELS[display.model].box_size

        # If the position of the window has been saved in the settings, restore
        # it.
        settings = QSettings("ledger", "speculos")
        # Take in account multiple screens and their geometry:
        current_screen_x = qt_app.primaryScreen().geometry().x()
        current_screen_y = qt_app.primaryScreen().geometry().y()
        if display.x is None:
            window_x = settings.value("window_x", current_screen_x + DEFAULT_WINDOW_X, int)
        else:
            window_x = display.x
        if display.y is None:
            window_y = settings.value("window_y", current_screen_y + DEFAULT_WINDOW_Y, int)
        else:
            window_y = display.y
        window_width = (self._width + box_size_x) * display.pixel_size
        window_height = (self._height + box_size_y) * display.pixel_size

        # Be sure Window is FULLY visible in one of the available screens:
        window_is_visible = False
        for screen in qt_app.screens():
            x1 = screen.geometry().x()
            y1 = screen.geometry().y()
            x2 = x1 + screen.geometry().width() - 1
            y2 = y1 + screen.geometry().height() - 1

            if window_x >= x1 and window_y >= y1 and (window_x + window_width - 1) <= x2 and \
               (window_y + window_height - 1) <= y2:
                window_is_visible = True
                break   # No need to check other screens

        # If the window is not FULLY visible, force default coordinates on current screen:
        if not window_is_visible:
            print("Window is NOT FULLY visible => using default coordinates in current screen.")
            window_x = current_screen_x + DEFAULT_WINDOW_X
            window_y = current_screen_y + DEFAULT_WINDOW_Y

        self.setGeometry(window_x, window_y, window_width, window_height)
        self.setFixedSize(window_width, window_height)

        if display.ontop:
            flags = Qt.CustomizeWindowHint | Qt.WindowStaysOnTopHint
            self.setWindowFlags(flags)

        self.setAutoFillBackground(True)
        p = self.palette()
        p.setColor(self.backgroundRole(), QColor.fromRgb(COLORS[display.color]))
        self.setPalette(p)

        # Add paint widget and paint
        self.widget = PaintWidget(self, display.model, display.pixel_size, server.vnc)
        self.widget.move(self.box_position_x * display.pixel_size, self.box_position_y * display.pixel_size)
        self.widget.resize(self._width * display.pixel_size, self._height * display.pixel_size)
        self.setWindowIcon(QIcon('mcu/icon.png'))
        self.show()
        self._screen: Screen

    def set_screen(self, screen: Screen) -> None:
        self._screen = screen

    def screen_update(self) -> bool:
        return self._screen.screen_update()

    def keyPressEvent(self, event: QKeyEvent):
        self._screen._key_event(event, True)

    def keyReleaseEvent(self, event: QKeyEvent):
        self._screen._key_event(event, False)

    def _get_x_y(self):
        x = self.mouse_offset.x() // self.pixel_size - (self.box_position_x + 1)
        y = self.mouse_offset.y() // self.pixel_size - (self.box_position_y + 1)
        return x, y

    def mousePressEvent(self, event: QMouseEvent):
        '''Get the mouse location.'''

        self.mouse_offset = event.pos()

        x, y = self._get_x_y()
        if x >= 0 and x < self._width and y >= 0 and y < self._height:
            self.seph.handle_finger(x, y, True)

        QApplication.setOverrideCursor(Qt.DragMoveCursor)

    def mouseReleaseEvent(self, event: QMouseEvent):
        self.mouse_offset = event.pos()
        x, y = self._get_x_y()
        if x >= 0 and x < self._width and y >= 0 and y < self._height:
            # Send the release
            self.seph.handle_finger(x, y, False)
        QApplication.restoreOverrideCursor()

    def closeEvent(self, event: QEvent):
        '''
        Called when the window is closed. We save the current window position to
        the settings file in order to restore it upon next speculos execution.
        '''
        settings = QSettings("ledger", "speculos")
        settings.setValue("window_x", self.pos().x())
        settings.setValue("window_y", self.pos().y())


class Screen(Display):
    def __init__(self, display: DisplayArgs, server: ServerArgs) -> None:
        super().__init__(display, server)
        self.app: App
        self._gl: GraphicLibrary

    def set_app(self, app: App) -> None:
        self.app = app
        self.app.set_screen(self)
        model = self._display_args.model
        if self.use_bagl:
            self._gl = bagl.Bagl(app.widget, MODELS[model].screen_size, model)
        else:
            self._gl = nbgl.NBGL(app.widget, MODELS[model].screen_size, model)

    @property
    def m(self) -> QWidget:
        return self.app.widget

    @property
    def gl(self):
        return self._gl

    def _key_event(self, event: QKeyEvent, pressed) -> None:
        key = Qt.Key(event.key())
        if key in [Qt.Key_Left, Qt.Key_Right]:
            buttons = {Qt.Key_Left: BUTTON_LEFT, Qt.Key_Right: BUTTON_RIGHT}
            # forward this event to seph
            self.seph.handle_button(buttons[key], pressed)
        elif key == Qt.Key_Down:
            self.seph.handle_button(BUTTON_LEFT, pressed)
            self.seph.handle_button(BUTTON_RIGHT, pressed)
        elif key == Qt.Key_Q and not pressed:
            self.app.close()

    def display_status(self, data: bytes) -> List[TextEvent]:
        assert isinstance(self.gl, bagl.Bagl)
        ret = self.gl.display_status(data)
        if MODELS[self.model].name == 'blue':
            self.screen_update()    # Actually, this method doesn't work
        return ret

    def display_raw_status(self, data: bytes) -> None:
        assert isinstance(self.gl, bagl.Bagl)
        self.gl.display_raw_status(data)
        if MODELS[self.model].name == 'blue':
            self.screen_update()    # Actually, this method doesn't work

    def screen_update(self) -> bool:
        return self.gl.refresh()


class QtScreenNotifier(DisplayNotifier):
    def __init__(self, display_args: DisplayArgs, server_args: ServerArgs) -> None:
        self._qapp = QApplication([])
        super().__init__(display_args, server_args)
        self._set_display_class(Screen)
        self._app_widget = App(self._qapp, display_args, server_args)
        assert isinstance(self.display, Screen)
        self.display.set_app(self._app_widget)

    def _can_read(self, device: IODevice) -> None:
        try:
            device.can_read(self)
        # This exception occur when can_read have no more data available
        except ReadError:
            self._app_widget.close()

    def add_notifier(self, device: IODevice) -> None:
        n = QSocketNotifier(voidptr(device.fileno), QSocketNotifier.Read, self._qapp)
        n.activated.connect(lambda _: self._can_read(device))
        assert device.fileno not in self.notifiers
        self.notifiers[device.fileno] = n

    def enable_notifier(self, fd: int, enabled: bool = True) -> None:
        n = self.notifiers[fd]
        n.setEnabled(enabled)

    def remove_notifier(self, fd: int) -> None:
        # just in case
        self.enable_notifier(fd, False)
        n = self.notifiers.pop(fd)
        n.disconnect()

    def run(self):
        self._qapp.exec_()
        self._qapp.quit()



import pytest
import os
import re
from collections import namedtuple
from pathlib import Path
from typing import List

from speculos.client import SpeculosClient


# going back from...(conftest.py \ python \ tests \ git root) / 'apps'
APP_DIR = Path(__file__).resolve().parent.  parent. parent    / "apps"
AppInfo = namedtuple("AppInfo", ["filepath", "model", "name", "sdk", "hash"])


def app_info_from_path(path: Path) -> AppInfo:
    # name example: nanos#btc#1.5#5b6693b8.elf
    app_regexp = re.compile(r"^(nanos|nanox|blue|nanosp)#([^#]+)#([^#][\d\w\-.]+)#([a-f0-9]*)\.elf$")
    matching = re.match(app_regexp, path.name)
    if not matching:
        return None
    assert len(matching.groups()) == 4
    return AppInfo(filepath=path, model=matching.group(1), name=matching.group(2), sdk=matching.group(3),
                   hash=matching.group(4))


def list_apps_to_test() -> List[AppInfo]:
    """
    List apps matching the pattern:

        <device>#<app_name>#<sdk_version>#<commit_hash>.elf

    in the apps/ directory and return a list of APDUClient
    objects for these applications.

    A typical application path looks like:

    'apps/nanos#btc#1.5#5b6693b8.elf'
    """
    all_apps = []
    for appfile in APP_DIR.iterdir():
        if "#" not in appfile.name:
            continue
        info = app_info_from_path(appfile)
        if not info:
            pytest.fail(
                f"An unexpected file was found in apps/, with a # but not matching the pattern: {filename!r}"
            )
            continue
        all_apps.append(info)
    return all_apps


@pytest.fixture(scope="function")
def app(request, client):
    return app_info_from_path(Path(client.app))


def get_apps(name: str) -> List[AppInfo]:
    """Retrieve the list of apps in the ../../apps directory."""
    return [app for app in list_apps_to_test() if app.name == name]


def default_btc_app() -> List[AppInfo]:
    filepath = (APP_DIR / "btc.elf").resolve()
    apps = get_apps("btc")
    return [app for app in apps if app.filepath == filepath]


def idfn(app: Path) -> str:
    """
    Set the test ID to the app file name for each test running on a set of apps.

    From https://docs.pytest.org/en/6.2.x/example/parametrize.html#different-options-for-test-ids:

      These IDs can be used with -k to select specific cases to run, and they will
      also identify the specific case when one is failing.
    """
    return app.filepath


def client_instance(app, additional_args=None):
    args = ["--model", app.model, "--sdk", app.sdk]
    if additional_args is not None:
        args += additional_args
    return SpeculosClient(str(app.filepath), args=args)


@pytest.fixture(scope="module", params=get_apps("btc"), ids=idfn)
def client_btc(request):
    with client_instance(request.param) as _client:
        yield _client


@pytest.fixture(scope="module", params=get_apps("btc-test"), ids=idfn)
def client_btc_testnet(request):
    app = request.param
    btc_app = app.filepath.parent / app.filepath.name.replace("btc-test", "btc")
    assert btc_app.is_file()
    args = ["-l", "Bitcoin:%s" % str(btc_app)]

    with client_instance(request.param, additional_args=args) as _client:
        yield _client


@pytest.fixture(scope="module", params=get_apps("ram-page"), ids=idfn)
def client_ram_page(request):
    with client_instance(request.param) as _client:
        yield _client


@pytest.fixture(scope="function", params=default_btc_app(), ids=idfn)
def client_vnc(request):
    # Pytest has changed its API in version 4: https://github.com/pytest-dev/pytest/pull/4564
    if hasattr(request.node, "get_closest_marker"):
        get_closest_marker = request.node.get_closest_marker
    else:
        get_closest_marker = request.node.get_marker
    args = list(get_closest_marker("additional_args").args)
    with client_instance(request.param, args) as _client:
        yield _client


@pytest.fixture(scope="class")
def client(request):
    """Run the API tests on the default btc.elf app."""

    info = app_info_from_path((APP_DIR / "btc.elf").resolve())
    args = ["--model", info.model, "--sdk", info.sdk]
    with SpeculosClient(app=str(info.filepath), args=args) as _client:
        yield _client



import importlib.resources
import json
import os
import pytest
import re
import requests
import time
from collections import namedtuple

from speculos.client import SpeculosClient

API_URL = "http://127.0.0.1:5000"


@pytest.mark.usefixtures("client")
class TestApi:
    @staticmethod
    def get_automation_data(name):
        path = importlib.resources.files(__package__) / "resources" / name
        with open(path, "rb") as fp:
            data = fp.read()
        return data

    def test_automation_valid(self):
        data = TestApi.get_automation_data("automation.json")
        with requests.post(f"{API_URL}/automation", data=data) as response:
            assert response.status_code == 200
            assert response.json() == {}

    def test_automation_invalid_path(self):
        with requests.post(f"{API_URL}/automation", data=b"file:/etc/passwd") as response:
            assert response.status_code == 400

    def test_automation_invalid_json(self):
        with requests.post(f"{API_URL}/automation", data=b"x") as response:
            assert response.status_code == 400

    @staticmethod
    def press_button(button):
        data = json.dumps({"action": "press-and-release"}).encode()
        with requests.post(f"{API_URL}/button/{button}", data=data) as response:
            assert response.status_code == 200

    def test_button(self):
        for button in ["right", "left", "both"]:
            TestApi.press_button(button)

    def test_finger(self):
        data = json.dumps({"x": 0, "y": 0, "action": "press-and-release"}).encode()
        with requests.post(f"{API_URL}/finger", data=data) as response:
            assert response.status_code == 200

    def test_events(self):
        """
        Read a stream of events while pressing the button 'right' and left 3
        times.
        """

        def get_next_event(stream):
            """
            Return the next event.

            A subset of the event stream format is recognized by this function
            and the event is expected to be encoded in JSON.
            """

            data = b""
            while True:
                line = stream.raw.readline()
                if line == b"\n":
                    break
                assert line.startswith(b"data: ")
                data += line[6:]
            event = json.loads(data)
            assert isinstance(event, dict)
            return event

        def get_current_screen_content(session):
            return session.get(f"{API_URL}/events?currentscreenonly=true").content.decode("utf-8")

        with requests.Session() as r:
            with r.get(f"{API_URL}/events?stream=true", stream=True) as stream:
                assert stream.status_code == 200

                texts = [("Version", ".*"), ("About",), ("Quit",)]
                for i in range(0, 3):
                    TestApi.press_button("right")
                    for text in texts[i]:
                        event = get_next_event(stream)
                        assert re.match(text, event["text"])

                texts = [("About",), ("Version", ".*"), ("Bitcoin", "is ready")]
                for i in range(len(texts)):
                    TestApi.press_button("left")
                    for text in texts[i]:
                        event = get_next_event(stream)
                        assert re.match(text, event["text"])

            texts = [
                '{"events": [{"text": "Bitcoin", "x": 41, "y": 3, "w": 128, "h": 32, "clear": false}, '
                '{"text": "is ready", "x": 41, "y": 17, "w": 128, "h": 32, "clear": false}]}\n',
                '{"events": [{"text": "Version", "x": 43, "y": 3, "w": 79, "h": 32, "clear": false}, '
                '{"text": "2.0.1", "x": 52, "y": 17, "w": 70, "h": 32, "clear": false}]}\n',
                '{"events": [{"text": "About", "x": 47, "y": 19, "w": 81, "h": 32, "clear": false}]}\n'
            ]
            for text in texts:
                content = get_current_screen_content(r)
                assert content == text
                TestApi.press_button("right")

                # Wait for the screen to be updated and parsed
                while content == get_current_screen_content(r):
                    time.sleep(1)

            with r.get(f"{API_URL}/events") as response:
                assert json.loads(response.content)

            with r.delete(f"{API_URL}/events") as response:
                assert response.status_code == 200

            with r.get(f"{API_URL}/events") as response:
                assert json.loads(response.content) == {"events": []}

    def test_screenshot(self):
        with requests.get(f"{API_URL}/screenshot") as response:
            assert response.status_code == 200
            assert response.headers["Content-Type"] == "image/png"
            assert response.content.startswith(b"\x89PNG")

    def test_apdu(self):
        # Send GET_RANDOM APDU to get 16 bytes of random
        with requests.post(f"{API_URL}/apdu", json={"data": "e0c0000010"}) as response:
            assert response.status_code == 200
            data = bytes.fromhex(response.json()["data"])
            assert len(data) == 18 and data[-2:] == b"\x90\x00"

    def test_apdu_invalid_data(self):
        with requests.post(f"{API_URL}/apdu", json={"data": "xyz"}) as response:
            assert response.status_code == 400
        with requests.post(f"{API_URL}/apdu") as response:  # Missing data field
            assert response.status_code == 400



import importlib.resources
import json
import jsonschema
import os
import pytest

from speculos.mcu import automation


class TestAutomation:
    @staticmethod
    def get_json_path(name):
        path = importlib.resources.files(__package__) / "resources" / name
        return f"file:{path}"

    def test_valid_json(self):
        """Valid JSON complying with the schema."""

        automation.Automation('{"version": 1, "rules": []}')
        automation.Automation(TestAutomation.get_json_path("automation_valid.json"))

    def test_invalid_json(self):
        """Invalid JSON/schema testcases."""

        with pytest.raises(json.decoder.JSONDecodeError):
            assert automation.Automation("x")

        with pytest.raises(jsonschema.exceptions.ValidationError):
            assert automation.Automation("{}")

        names = [
            "automation_invalid_rule_key.json",
            "automation_invalid_action_name.json",
            "automation_invalid_action_args.json",
        ]
        for name in names:
            with pytest.raises(jsonschema.exceptions.ValidationError):
                assert automation.Automation(TestAutomation.get_json_path(name))

    def test_rules(self):
        expected_actions = [
            ["button", 2, True],
            ["button", 2, False],
            ["setbool", "seen", True],
            ["exit"]
        ]
        regexp_actions = [
            ["exit"]
        ]
        default_actions = [
            ["setbool", "default_match", True]
        ]

        auto = automation.Automation(TestAutomation.get_json_path("automation_valid.json"))
        assert auto.get_actions("Application", 0, 0) == default_actions
        assert auto.get_actions("Application", 35, 3) == expected_actions

        auto.set_bool("seen", True)
        assert auto.get_actions("Application", 35, 3) == default_actions

        auto.set_bool("seen", False)
        assert auto.get_actions("Application", 35, 3) == expected_actions

        assert auto.get_actions("1234", 35, 3) == regexp_actions



import functools
import string
from dataclasses import dataclass
from typing import Dict, List, Mapping
from speculos.observer import TextEvent
from . import bagl_font

from construct import Struct, Int8ul, Int16ul

nbgl_area_t = Struct(
    "x0" / Int16ul,
    "y0" / Int16ul,
    "width" / Int16ul,
    "height" / Int16ul,
    "color" / Int8ul,
    "bpp" / Int8ul,
)

MIN_WORD_CONFIDENCE_LVL = 0  # percent
NEW_LINE_THRESHOLD = 10  # pixels

BitMap = bytes
BitVector = str  # a string of '1' and '0'
Width = int  # width
CharWidth = Width
Char = str  # a single character string (chars are 1-char strings in Python)


@dataclass
class BitMapChar:
    char: bagl_font.FontCharacter
    bitmap: bytes


__FONT_MAP: Dict[int, Mapping[str, BitMapChar]] = {}

DISPLAY_CHARS = string.ascii_letters + string.digits + string.punctuation


def cache_font(f):
    __font_char_cache = {}

    @functools.wraps(f)
    def wrapper(byte_string: bytes):
        if byte_string not in __font_char_cache:
            __font_char_cache[byte_string] = f(byte_string)
        return __font_char_cache[byte_string]
    return wrapper


def split(bits: BitVector, n: Width) -> List[BitVector]:
    """
    Split a bit array (string of '1' and '0')
    into an arbitrary list of n-sized bit arrays.

    Useful if you want to display a character from a bitmap
    """
    output = []
    bit_list = bits[:]
    while bit_list:
        bit_list_to_append = bit_list[:n]
        # do padding
        # bit_list_to_append += '0' * (n - len(bit_list_to_append))
        output.append(bit_list_to_append)
        bit_list = bit_list[n:]
    return output


def split_bytes(bytes_elems: BitMap, n: CharWidth):
    """Useful to display a bitmap character"""
    bits = "".join("{0:08b}".format(x)[::-1] for x in bytes_elems)
    return split(bits, n)


def get_font(font_id: int) -> bagl_font.Font:
    font = bagl_font.get(font_id)
    if not font:
        raise ValueError(f"Could not get font with font_id={font_id}")
    return font


def get_char(font: bagl_font.Font, char: str) -> BitMapChar:
    if len(char) != 1 or char not in DISPLAY_CHARS:
        raise ValueError(f"could not print {char}")
    return get_font_map(font)[char]


def display_char(font: bagl_font.Font, char: str) -> None:
    bm_char = get_char(font, char)
    print("\n".join(split_bytes(bm_char.bitmap, font.bpp * bm_char.char.char_width)))


def get_font_map(font: bagl_font.Font):
    if font.font_id not in __FONT_MAP:
        __FONT_MAP[font.font_id] = _get_font_map(font)
    return __FONT_MAP[font.font_id]


def _get_font_map(font: bagl_font.Font) -> Mapping[Char, BitMapChar]:
    font_map = {}
    for ord_char, font_char in zip(
        range(font.first_char, font.last_char), font.characters
    ):
        font_map[chr(ord_char)] = BitMapChar(
            font_char,
            bytes(
                font.bitmap[
                    font_char.bitmap_offset:(
                        font_char.bitmap_offset
                        + font_char.bitmap_byte_count
                    )
                ]
            ),
        )
    return font_map


class OCR:

    # Maximum space for a letter to be considered part of the same word
    MAX_BLANK_SPACE_NANO = 12
    MAX_BLANK_SPACE_STAX = 24
    MAX_BLANK_SPACE_FLEX = 26

    def __init__(self, model: str, use_bagl: bool):
        self.events: List[TextEvent] = []
        # Store the model of the device
        self.model = model
        self.use_bagl = use_bagl
        # Maximum space for a letter to be considered part of the same word
        if model == "stax":
            self.max_blank_space = OCR.MAX_BLANK_SPACE_STAX
        elif model == "flex":
            self.max_blank_space = OCR.MAX_BLANK_SPACE_FLEX
        else:
            self.max_blank_space = OCR.MAX_BLANK_SPACE_NANO

    @staticmethod
    def find_char_from_bitmap(bitmap: BitMap) -> str:
        """
        Find a character from a bitmap
        >>> font = get_font(4)
        >>> char = get_char(font, 'c')
        >>> find_char_from_bitmap(char.bitmap)
        'c'
        """
        all_values = []
        for font in bagl_font.FONTS:
            font_map = get_font_map(font)
            for character_value, bitmap_struct in font_map.items():
                if bitmap_struct.bitmap.startswith(bitmap):
                    # sometimes (but not always) the bitmap being passed is shortened
                    # by one '\x00' byte, not matching the exact bitmap
                    # provided in the font. Hence the 'residual' computation
                    residual_bytes: bytes = bitmap_struct.bitmap[len(bitmap):]
                    if all(b == 0 for b in residual_bytes):
                        all_values.append(character_value)
            if all_values:
                char = max([x for x in all_values])
                if char == "\x80":
                    char = " "
                return char
        return ""

    def find_bitmap(self, x: int, y: int, w: int, h: int, bitmap: bytes) -> None:
        char = self.find_char_from_bitmap(bitmap)
        if char:
            if self.events and y <= self.events[-1].y:
                self.events[-1].text += char
            else:
                # create a new TextEvent if there are no events yet
                # or if there is a new line
                self.events.append(TextEvent(char, x, y, w, h, False))

    def store_char_in_last_event(self, x: int, y: int, w: int, h: int, char: str) -> None:
        """
        Add current character to last event
        """
        self.events[-1].text += char
        # Update w for all chars in self.events[-1]
        x2 = x + w - 1
        self.events[-1].w = x2 - self.events[-1].x + 1
        # Update y & h, if needed, for all chars in self.events[-1]
        y1 = y
        if y1 > self.events[-1].y:
            # Keep the lowest Y in Y1
            y1 = self.events[-1].y
        y2 = y + h - 1
        if y2 < (self.events[-1].y + self.events[-1].h):
            # Keep the highest Y in Y2
            y2 = self.events[-1].y + self.events[-1].h - 1
        self.events[-1].y = y1
        self.events[-1].h = y2 - y1 + 1

    def add_character(self, x: int, y: int, w: int, h: int, char: str) -> None:
        """
        Add the provided character to previous event or create a new one
        """
        # Compute difference with X coord from previous event
        # if x_diff > self.max_blank_space the char is not on same sentence
        if self.events:
            x_diff = x - (self.events[-1].x + self.events[-1].w)
            if x_diff < 0:
                x_diff = -x_diff
            # Try to find if that char can be added to previous event
            if y < (self.events[-1].y + self.events[-1].h) \
               and x_diff < self.max_blank_space:
                # Add this character to previous event
                self.store_char_in_last_event(x, y, w, h, char)
                return

        # create a new TextEvent if there are no events yet or if there is a new line
        self.events.append(TextEvent(char, x, y, w, h, False))

    def analyze_bitmap(self, data: bytes) -> None:
        """
        data contain information about the latest displayed bitmap.
        Since unified SDK, speculos added the displayed character to data.
        For older SKD versions, legacy behaviour is used: parsing internal
        fonts to find a matching bitmap.
        """
        if not self.use_bagl:
            # Can be called via SephTag.NBGL_DRAW_IMAGE or SephTag.NBGL_DRAW_IMAGE_RLE
            # In both cases, data contains:
            # - area (sizeof(nbgl_area_t))
            # - compressed bitmap (buffer_len)
            # - 2 bytes of different meaning depending on SephTag
            # - 4 bytes with unicode code point of displayed character
            area = nbgl_area_t.parse(data[0:nbgl_area_t.sizeof()])
            x, y, w, h = area.x0, area.y0, area.width, area.height
            character = int.from_bytes(data[-4:], byteorder="little", signed=False)
            bitmap = data[nbgl_area_t.sizeof():-(2+4)]
        else:
            if data[0] != 0:
                return

            x = int.from_bytes(data[1:3], byteorder="big", signed=True)
            y = int.from_bytes(data[3:5], byteorder="big", signed=True)
            w = int.from_bytes(data[5:7], byteorder="big", signed=True)
            h = int.from_bytes(data[7:9], byteorder="big", signed=True)
            bpp = int.from_bytes(data[9:10], byteorder="big")
            character = int.from_bytes(data[10:14], byteorder="big", signed=False)
            color_size = 4 * (1 << bpp)
            bitmap = data[14+color_size:]

            # h may no reflect the real height: use number of lines displayed
            h = (len(bitmap) * 8) // w
            if (len(bitmap) * 8) % w:
                h += 1

        # Space is now encoded as an empty character (no 'space' wasted :)
        if len(bitmap) == 0:
            character = 32

        # If data contains a character, don't bother doing OCR, just store it!
        if character != 0:
            char = chr(character)
            self.add_character(x, y, w, h, char)
        else:
            self.find_bitmap(x, y, w, h, bitmap)

    def get_events(self) -> List[TextEvent]:
        events = self.events.copy()
        self.events = []
        return events



import jsonschema
from flask import request

from speculos.resources_importer import get_resource_schema_as_json
from .restful import SephResource


class Ticker(SephResource):
    schema = get_resource_schema_as_json("api", "ticker.schema")

    def post(self):
        args = request.get_json(force=True)
        try:
            jsonschema.validate(instance=args, schema=self.schema)
        except jsonschema.exceptions.ValidationError as e:
            return {"error": f"{e}"}, 400

        action = args["action"]
        self.seph.handle_ticker_request(action)
        return {}, 200



"""
Forward screen events (text, x, y) to TCP clients.
"""

import json
import logging
import socketserver
import threading
from typing import List
from dataclasses import asdict

from speculos.observer import BroadcastInterface, ObserverInterface, TextEvent


class AutomationServer(socketserver.ThreadingMixIn, socketserver.TCPServer, BroadcastInterface):
    daemon_threads = True
    allow_reuse_address = True

    def __init__(self, server_address, RequestHandlerClass, *args, **kwargs):
        BroadcastInterface.__init__(self)
        socketserver.TCPServer.__init__(self, server_address, RequestHandlerClass)
        self.logger = logging.getLogger("automation")

    def broadcast(self, event: TextEvent):
        """Broadcast an event to each connected client."""

        self.logger.debug("Broadcast %s to %s", asdict(event), self.clients)
        for client in self.clients:
            client.send_screen_event(event)


class AutomationClient(socketserver.BaseRequestHandler, ObserverInterface):

    def setup(self) -> None:
        self.server:  AutomationServer
        self.logger = logging.getLogger("automation")
        self.condition = threading.Condition()
        self.events: List[TextEvent] = []
        self.logger.debug("New client from %s", self.client_address)
        self.server.add_client(self)

    def handle(self):
        while True:
            with self.condition:
                while len(self.events) == 0:
                    self.condition.wait()

            while self.events:
                event = self.events.pop(0)
                data = json.dumps(asdict(event)).encode()
                try:
                    self.request.sendall(data + b"\n")
                except BrokenPipeError:
                    return

    def finish(self):
        self.logger.debug("Connection closed with client %s", self)
        self.server.remove_client(self)

    def send_screen_event(self, event: TextEvent) -> None:
        self.events.append(event)
        with self.condition:
            self.condition.notify()



#!/usr/bin/python2

'''
VNC client with skin support.

Speculos must be launched with the --vnc-port argument in order to expose the
VNC server. Usage example:

  $ ./speculos.py ./apps/nanos#btc#1.5#00000000.elf --vnc-port 5902 --headless &
  $ ./tools/winamp.py --skin tools/skins/nano-blue.png

This script is highly experimental. Skins are PNG files where the first black
128x32 rectangle is replaced by the screen.

Requirements: python-gtk-vnc. Unfortunately, there seems to be no Python 3
version of gtkvnc.

This script is based on the example from
https://github.com/jwendell/gtk-vnc/blob/master/examples/gvncviewer.py
'''

import argparse
import gtk
import gtkvnc
import time
import cairo

from gtk import gdk
from PIL import Image

OFFSET_X, OFFSET_Y = 0, 0
SCREEN_WIDTH, SCREEN_HEIGHT = 128, 32


def button_press_event(box, event):
    global OFFSET_X, OFFSET_Y

    if event.button == 1:
        window = box.get_parent()
        x, y = window.get_position()
        OFFSET_X = event.x_root - x
        OFFSET_Y = event.y_root - y

        arrow = gdk.Cursor(gdk.FLEUR)
        root = window.get_root_window()
        root.set_cursor(arrow)


def button_release_event(box, event):
    if event.button == 1:
        arrow = gdk.Cursor(gdk.LEFT_PTR)
        root = window.get_root_window()
        root.set_cursor(arrow)


def motion_notify_event(box, event):
    global OFFSET_X, OFFSET_Y

    window = box.get_parent()
    x = int(event.x_root - OFFSET_X)
    y = int(event.y_root - OFFSET_Y)
    window.move(x, y)


def vnc_screenshot(src, ev, vnc):
    if ev.keyval == gtk.gdk.keyval_from_name("s"):
        filename = "/tmp/speculos-{}.png".format(int(time.time()))
        pix = vnc.get_pixbuf()
        pix.save(filename, "png")
        print("[*] screenshot saved to {}".format(filename))

    return False


def vnc_grab(src, window):
    pass


def vnc_ungrab(src, window):
    pass


def vnc_connected(src):
    print("Connected to server")


def vnc_initialized(src, window):
    print("Connection initialized")
    window.show_all()


def vnc_disconnected(src):
    print("Disconnected from server")
    gtk.main_quit()


def expose(widget, event):
    """
    https://stackoverflow.com/questions/4889045/how-to-get-transparent-background-in-window-with-pygtk-and-pycairo
    """

    window = widget.window.get_parent()
    cr = window.cairo_create()

    # Sets the operator to clear which deletes everything below where an object is drawn
    cr.set_operator(cairo.OPERATOR_CLEAR)
    cr.rectangle(0.0, 0.0, IMG_WIDTH, IMG_HEIGHT)
    cr.fill()

    cr.set_operator(cairo.OPERATOR_OVER)
    img = cairo.ImageSurface.create_from_png(IMG_PATH)
    cr.set_source_surface(img, 0, 0)
    cr.rectangle(0.0, 0.0, IMG_WIDTH, IMG_HEIGHT)
    cr.paint()
    cr.fill()


def is_black_rectangle(pixels, x, y, width, height):
    black = (0, 0, 0, 255)
    for i in range(0, SCREEN_WIDTH):
        for j in range(0, SCREEN_HEIGHT):
            if pixels[x+i, y+j] != black:
                return False
    return True


def find_screen_position(image):
    width, height = image.size
    pixels = image.load()
    for x in range(0, width - SCREEN_WIDTH):
        for y in range(0, height - SCREEN_HEIGHT):
            if is_black_rectangle(pixels, x, y, SCREEN_WIDTH, SCREEN_HEIGHT):
                return x, y

    # if not found, place the screen at the center of the image
    print("[-] can't found the screen position in the skin")
    x = (width - SCREEN_WIDTH) / 2
    y = (height - SCREEN_HEIGHT) / 2
    return x, y


if __name__ == '__main__':
    parser = argparse.ArgumentParser()
    parser.add_argument('-o', '--host', default='127.0.0.1', help='VNC host')
    parser.add_argument('-p', '--port', default='5902', help='VNC port')
    parser.add_argument('-s', '--skin', default=None, help='PNG image')
    args = parser.parse_args()

    window = gtk.Window(gtk.WINDOW_TOPLEVEL)
    vnc = gtkvnc.Display()

    if args.skin:
        image = Image.open(args.skin)
        IMG_PATH = args.skin
        IMG_WIDTH, IMG_HEIGHT = image.size
        x, y = find_screen_position(image)
        image.close()

        window.set_default_size(IMG_WIDTH, IMG_HEIGHT)
        fix = gtk.Fixed()
        fix.put(vnc, x, y)
        box = gtk.EventBox()
        window.add(box)
        box.add(fix)
    else:
        window.add(vnc)
        box = window

    box.connect("button_press_event", button_press_event)
    box.connect("button_release_event", button_release_event)
    box.connect("motion_notify_event", motion_notify_event)
    window.connect("key-press-event", vnc_screenshot, vnc)

    # borderless
    window.set_decorated(False)

    # transparent background
    screen = window.get_screen()
    rgba = screen.get_rgba_colormap()
    window.set_colormap(rgba)

    vnc.realize()
    vnc.set_pointer_grab(False)
    vnc.set_keyboard_grab(True)

    # Example to change grab key combination to Ctrl+Alt+g
    grab_keys = [gtk.keysyms.Control_L, gtk.keysyms.Alt_L, gtk.keysyms.g]
    vnc.set_grab_keys(grab_keys)

    print("Connecting to %s %s" % (args.host, args.port))
    vnc.open_host(args.host, args.port)

    vnc.connect("vnc-pointer-grab", vnc_grab, window)
    vnc.connect("vnc-pointer-ungrab", vnc_ungrab, window)

    vnc.connect("vnc-connected", vnc_connected)
    vnc.connect("vnc-initialized", vnc_initialized, window)
    vnc.connect("vnc-disconnected", vnc_disconnected)

    if args.skin:
        vnc.connect("expose-event", expose)

    gtk.main()



from abc import ABC, abstractmethod
from dataclasses import dataclass
from logging import Logger
from typing import List


@dataclass
class TextEvent:
    text: str
    x: int
    y: int
    w: int
    h: int
    clear: bool


class ObserverInterface(ABC):

    @abstractmethod
    def send_screen_event(self, event: TextEvent) -> None:
        pass


class BroadcastInterface(ABC):

    def __init__(self) -> None:
        self.logger: Logger
        self.clients: List[ObserverInterface] = list()

    def add_client(self, client: ObserverInterface) -> None:
        self.logger.debug("New client '%s'", client)
        self.clients.append(client)

    def remove_client(self, client: ObserverInterface) -> None:
        self.logger.debug("Removing client '%s'", client)
        self.clients.remove(client)

    @abstractmethod
    def broadcast(self, event: TextEvent) -> None:
        pass



import select
from typing import List, Optional

from speculos.observer import TextEvent
from . import bagl
from . import nbgl
from .display import Display, DisplayNotifier, FrameBuffer, GraphicLibrary, MODELS
from .readerror import ReadError
from .struct import DisplayArgs, ServerArgs
from .vnc import VNC


class Headless(Display):
    def __init__(self, display: DisplayArgs, server: ServerArgs) -> None:
        super().__init__(display, server)

        self.m = HeadlessPaintWidget(self.model, server.vnc)
        self._gl: GraphicLibrary
        if display.use_bagl:
            self._gl = bagl.Bagl(self.m, MODELS[self.model].screen_size, self.model)
        else:
            self._gl = nbgl.NBGL(self.m, MODELS[self.model].screen_size, self.model)

    @property
    def gl(self) -> GraphicLibrary:
        return self._gl

    def display_status(self, data: bytes) -> List[TextEvent]:
        assert isinstance(self.gl, bagl.Bagl)
        ret = self.gl.display_status(data)
        if MODELS[self.model].name == 'blue':
            self.screen_update()    # Actually, this method doesn't work
        return ret

    def display_raw_status(self, data: bytes) -> None:
        assert isinstance(self.gl, bagl.Bagl)
        self.gl.display_raw_status(data)
        if MODELS[self.model].name == 'blue':
            self.screen_update()    # Actually, this method doesn't work

    def screen_update(self) -> bool:
        assert isinstance(self.gl, bagl.Bagl)
        return self.gl.refresh()


class HeadlessPaintWidget(FrameBuffer):
    def __init__(self, model: str, vnc: Optional[VNC] = None):
        super().__init__(model)
        self.vnc = vnc

    def update(self,
               _0: Optional[int] = None,
               _1: Optional[int] = None,
               _2: Optional[int] = None,
               _3: Optional[int] = None) -> bool:
        if self.pixels or self.draw_default_color:
            self._redraw()
            self.pixels = {}
            self.draw_default_color = False
            return True
        return False

    def _redraw(self) -> None:
        if self.vnc:
            self.vnc.redraw(self.pixels, self.default_color)
        self.update_screenshot()


class HeadlessNotifier(DisplayNotifier):

    def __init__(self, display_args: DisplayArgs, server_args: ServerArgs) -> None:
        super().__init__(display_args, server_args)
        self._set_display_class(Headless)

    def run(self) -> None:
        while True:
            _rlist = self.notifiers.keys()
            if not _rlist:
                break

            rlist, _, _ = select.select(_rlist, [], [])
            try:
                for fd in rlist:
                    self.notifiers[fd].can_read(self)

            # This exception occur when can_read have no more data available
            except ReadError:
                break



import jsonschema
from flask import request

from speculos.resources_importer import get_resource_schema_as_json
from .restful import SephResource


class Finger(SephResource):
    schema = get_resource_schema_as_json("api", "finger.schema")

    def post(self):
        args = request.get_json(force=True)
        try:
            jsonschema.validate(instance=args, schema=self.schema)
        except jsonschema.exceptions.ValidationError as e:
            return {"error": f"{e}"}, 400

        action = args["action"]
        delay = args.get("delay", 0.1)

        x1, y1 = args["x"], args["y"]
        if action == "press-and-release":
            self.seph.handle_finger(x1, y1, True)
            self.seph.handle_wait(delay)
            x2, y2 = args.get("x2", x1), args.get("y2", y1)
            self.seph.handle_finger(x2, y2, False)
        else:
            self.seph.handle_finger(x1, y1, action == "press")

        return {}, 200



#!/usr/bin/env python3

'''
Tests to ensure that speculos launches correctly the BTC apps.
'''

import json
import importlib.resources
import io
import os
import pytest

from enum import IntEnum

import speculos.client

CLA = 0xE0


class Ins(IntEnum):
    GET_PUBLIC_KEY = 0x40
    GET_VERSION = 0xC4


@pytest.fixture(scope="module")
def client(client_btc):
    return client_btc


def read_automation_rules(name):
    path = importlib.resources.files(__package__) / "resources" / name
    with open(path, "rb") as fp:
        rules = json.load(fp)
    return rules


def test_btc_get_version(client):
    """Send a get_version APDU to the BTC app."""
    client.apdu_exchange(CLA, Ins.GET_VERSION, b"")


def test_btc_get_public_key_with_user_approval(client, app):
    """Ask for the public key, compare screenshots and press reject."""

    if app.model == "nanos" and app.hash == "00000000":
        pytest.skip(f"skipped on model {app.model} because of animated text")

    bip32_path = bytes.fromhex("8000002C80000000800000000000000000000000")
    payload = bytes([len(bip32_path) // 4]) + bip32_path

    with client.apdu_exchange_nowait(CLA, Ins.GET_PUBLIC_KEY, payload, p1=0x01) as response:
        if app.model == "blue":
            client.wait_for_text_event("CONFIRM ACCOUNT")
        else:
            event = client.wait_for_text_event("Address")
            while "Reject" not in event["text"]:
                client.press_and_release("right")
                event = client.get_next_event()

        screenshot = client.get_screenshot()
        path = importlib.resources.files(__package__) / "resources" / f"btc_getpubkey_{app.model}.png"
        assert speculos.client.screenshot_equal(path, io.BytesIO(screenshot))

        if app.model == "blue":
            client.finger_touch(110, 440)
        else:
            client.press_and_release("both")

        with pytest.raises(speculos.client.ApduException) as e:
            response.receive()
        assert e.value.sw == 0x6985


def test_btc_automation(client, app):
    """Retrieve the pubkey, which requires a validation"""

    if app.model == "nanos" and app.hash == "00000000":
        pytest.skip(f"skipped on model {app.model} and revision {app.hash}")

    rules = read_automation_rules(f"btc_getpubkey_{app.model}.json")
    client.set_automation_rules(rules)

    payload = bytes.fromhex("058000003180000000800000000000000000000000")
    client.apdu_exchange(CLA, Ins.GET_PUBLIC_KEY, payload, p1=0x01, p2=0x01)



import socket
import threading
from typing import Any, Dict
from flask import Flask
from flask_restful import Api

from speculos.mcu.display import DisplayNotifier, IODevice
from speculos.mcu.readerror import ReadError
from speculos.mcu.seproxyhal import SeProxyHal
from speculos.observer import BroadcastInterface
from speculos.resources_importer import resources
from .apdu import APDU
from .automation import Automation
from .button import Button
from .events import Events
from .finger import Finger
from .screenshot import Screenshot
from .swagger import Swagger
from .web_interface import WebInterface
from .ticker import Ticker


class ApiRunner(IODevice):
    """Run the Speculos API server in a dedicated thread, with a notification when it stops"""
    def __init__(self, api_port: int) -> None:
        self._api_wrapper: ApiWrapper
        # self.sock is used by Screen.add_notifier. Closing self._notify_exit
        # signals it that the API is no longer running.
        self.sock: socket.socket
        self._notify_exit: socket.socket
        self.sock, self._notify_exit = socket.socketpair()
        self._port: int = api_port
        self._api_thread: threading.Thread

    @property
    def file(self):
        return self.sock

    def can_read(self, screen: DisplayNotifier) -> None:
        # Being able to read from the socket only happens when the API server exited.
        raise ReadError("API server exited")

    def start_server_thread(self,
                            screen_: DisplayNotifier,
                            seph_: SeProxyHal,
                            automation_server: BroadcastInterface) -> None:
        self._api_wrapper = ApiWrapper(self._port, screen_, seph_, automation_server)
        self._api_thread = threading.Thread(target=self._api_wrapper.run, name="API-server", daemon=True)
        self._api_thread.start()

    def stop(self):
        self._notify_exit.close()


class ApiWrapper:
    def __init__(self,
                 api_port: int,
                 screen: DisplayNotifier,
                 seph: SeProxyHal,
                 automation_server: BroadcastInterface):
        self._port = api_port
        static_folder = str(resources.files(__package__) / "static")
        self._app = Flask(__name__, static_url_path="", static_folder=static_folder)
        self._app.env = "development"

        screen_kwargs = {"screen": screen}
        seph_kwargs = {"seph": seph}
        app_kwargs = {"app": self._app}
        event_kwargs: Dict[str, Any] = {**app_kwargs, "automation_server": automation_server}

        self._api = Api(self._app)

        self._api.add_resource(APDU, "/apdu", resource_class_kwargs=seph_kwargs)
        self._api.add_resource(Automation, "/automation", resource_class_kwargs=seph_kwargs)
        self._api.add_resource(Button,
                               "/button/left",
                               "/button/right",
                               "/button/both",
                               resource_class_kwargs=seph_kwargs)
        self._api.add_resource(Events, "/events", resource_class_kwargs=event_kwargs)
        self._api.add_resource(Finger, "/finger", resource_class_kwargs=seph_kwargs)
        self._api.add_resource(Screenshot, "/screenshot", resource_class_kwargs=screen_kwargs)
        self._api.add_resource(Swagger, "/swagger/", resource_class_kwargs=app_kwargs)
        self._api.add_resource(WebInterface, "/", resource_class_kwargs=app_kwargs)
        self._api.add_resource(Ticker, "/ticker/", resource_class_kwargs=seph_kwargs)

    def run(self):
        # threaded must be set to allow serving requests along events streaming
        self._app.run(host="0.0.0.0", port=self._port, threaded=True, use_reloader=False)



#!/usr/bin/env python3

'''
Quick and dirty code to exchange APDUs between Ledger Live Speculos APDU server.

Usage example:

  export DEBUG_COMM_HTTP_PROXY='http://127.0.0.1:9998'
  ledger-live getAddress -c btc --path "m/49'/0'/0'/0/0" --derivationMode segwit
'''

import argparse
import binascii
import json
import http.server
import socketserver
import socket
import sys

HOST = '127.0.0.1'
PORT = 9998


def _recvall(s, size):
    data = b''
    while size > 0:
        tmp = s.recv(size)
        if len(tmp) == 0:
            print("[-] apduthread: connection with client closed", file=sys.stderr)
            return None
        data += tmp
        size -= len(tmp)
    return data


class SimpleHTTPRequestHandler(http.server.BaseHTTPRequestHandler):

    def do_GET(self):
        self.send_response(200)
        self.send_header("Access-Control-Allow-Origin", "*")
        self.end_headers()

    def do_OPTIONS(self):
        self.send_response(200)
        self.send_header("Access-Control-Allow-Origin", self.headers["Origin"])
        self.send_header("Access-Control-Allow-Headers", "Origin, X-Requested-With, Content-Type, Accept")
        self.send_header("Access-Control-Allow-Methods", "GET, POST")
        self.end_headers()

    def do_POST(self):
        # get APDU from the request body
        content_length = int(self.headers['Content-Length'])
        body = self.rfile.read(content_length)
        apdu = json.loads(body)
        apdu = binascii.unhexlify(apdu['apduHex'])

        with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as s:
            s.connect(('127.0.0.1', int(args.port)))

            if args.verbose:
                print('<', binascii.hexlify(apdu))

            # forward the APDU to the APDU server
            apdu = len(apdu).to_bytes(4, 'big') + apdu
            s.sendall(apdu)

            # receive the APDU server response
            data = _recvall(s, 4)
            size = int.from_bytes(data, byteorder='big')
            packet = _recvall(s, size + 2)

            if args.verbose:
                print('>', binascii.hexlify(packet))

        # forward the APDU response to the client, embedded in response body
        data = {'data': str(binascii.hexlify(packet), 'ascii'), 'error': None}
        data = json.dumps(data)
        data = data.encode('ascii')

        self.send_response(200)
        self.send_header('Content-Length', len(data))
        self.send_header('Content-Type', 'application/json')
        self.send_header("Access-Control-Allow-Origin", "*")
        self.end_headers()
        self.wfile.write(data)


if __name__ == '__main__':
    parser = argparse.ArgumentParser(description='Run one or more QEMU instances given the config files passed as '
                                                 'arguments. The -c/--config argument can be passed several times '
                                                 '(please note that the arguments order matters.)')
    parser.add_argument('-p', '--port', default=9999, help='APDU server port')
    parser.add_argument('-v', '--verbose', action='store_true')
    args = parser.parse_args()

    socketserver.TCPServer.allow_reuse_address = True
    with socketserver.TCPServer((HOST, PORT), SimpleHTTPRequestHandler) as httpd:
        print("serving at port", PORT)
        httpd.serve_forever()



#!/usr/bin/env python3

'''
Emulate the target app along the SE Proxy Hal server.
'''

import argparse
import binascii
import ctypes
import logging
import os
import re
import signal
import socket
import sys
import threading
from elftools.elf.elffile import ELFFile
from ledgered.binary import LedgerBinaryApp
from mnemonic import mnemonic
from typing import Optional, Type

from .api import ApiRunner, EventsBroadcaster
from .mcu import apdu as apdu_server
from .mcu import automation
from .mcu import display
from .mcu import seproxyhal
from .mcu.automation_server import AutomationClient, AutomationServer
from .mcu.button_tcp import FakeButton
from .mcu.finger_tcp import FakeFinger
from .mcu.struct import DisplayArgs, ServerArgs
from .mcu.vnc import VNC
from .observer import BroadcastInterface
from .resources_importer import resources


DEFAULT_SEED = ('glory promote mansion idle axis finger extra february uncover one trip resource lawn turtle enact '
                'monster seven myth punch hobby comfort wild raise skin')

logger = logging.getLogger("speculos")

launcher_path = str(resources.files(__package__) / "resources" / "launcher")


def set_pdeath(sig):
    '''Set the parent death signal of the calling process.'''

    PR_SET_PDEATHSIG = 1
    libc = ctypes.cdll.LoadLibrary('libc.so.6')
    libc.prctl(PR_SET_PDEATHSIG, sig)


def get_elf_infos(app_path, use_bagl):
    with open(app_path, 'rb') as fp:
        elf = ELFFile(fp)
        text = elf.get_section_by_name('.text')
        for seg in elf.iter_segments():
            if seg['p_type'] != 'PT_LOAD':
                continue
            if seg.section_in_segment(text):
                text_seg = seg
                break
        else:
            raise RuntimeError("No program header with text section!")
        symtab = elf.get_section_by_name('.symtab')
        bss = elf.get_section_by_name('.bss')
        sh_offset = text_seg['p_offset']
        sh_size = text_seg['p_filesz']
        text_load_addr = text['sh_addr']
        stack = bss['sh_addr']
        sym_estack = symtab.get_symbol_by_name('_estack')
        if sym_estack is None:
            sym_estack = symtab.get_symbol_by_name('END_STACK')
        if sym_estack is None:
            logger.error('failed to find _estack/END_STACK symbol')
            sys.exit(1)
        estack = sym_estack[0]['st_value']

        # Look for the symbols SVC_Call and SVC_cx_call
        # if they are found, save their addresses to patch them to replace the SYSCALL later
        svc_call_addr = 0
        svc_cx_call_addr = 0
        svc_call_symbol = symtab.get_symbol_by_name("SVC_Call")
        if svc_call_symbol is not None:
            svc_call_addr = svc_call_symbol[0]['st_value'] & (~1)
        svc_cx_call_symbol = symtab.get_symbol_by_name("SVC_cx_call")
        if svc_cx_call_symbol is not None:
            svc_cx_call_addr = svc_cx_call_symbol[0]['st_value'] & (~1)
        # Check where are located fonts in .elf file (LNX/LNS+ with BAGL only)
        # (on apps using NBGL, fonts are loaded from a known location: STAX_FONTS_ARRAY_ADDR,
        #  FLEX_FONTS_ARRAY_ADDR, NANOX_FONTS_ARRAY_ADDR or NANOSP_FONTS_ARRAY_ADDR)
        fonts_addr = 0
        fonts_size = 0
        bagl_fonts_symbol = symtab.get_symbol_by_name('C_bagl_fonts')
        if bagl_fonts_symbol is not None:
            fonts_addr = bagl_fonts_symbol[0]['st_value']
            fonts_size = bagl_fonts_symbol[0]['st_size']
            logger.info(f"Found C_bagl_fonts at 0x{fonts_addr:X} ({fonts_size} bytes)\n")
        elif use_bagl:
            logger.info("Disabling OCR.")

        supp_ram = elf.get_section_by_name('.rfbss')
        ram_addr, ram_size = (supp_ram['sh_addr'], supp_ram['sh_size']) if supp_ram is not None else (0, 0)
    stack_size = estack - stack
    return sh_offset, sh_size, stack, stack_size, ram_addr, ram_size, text_load_addr, \
        svc_call_addr, svc_cx_call_addr, fonts_addr, fonts_size


def get_cx_infos(app_path):
    with open(app_path, 'rb') as fp:
        elf = ELFFile(fp)
        text = elf.get_section_by_name('.text')
        cxram = elf.get_section_by_name('.cxram')
        sh_offset = text['sh_offset']
        sh_size = text['sh_size']
        sh_load = text['sh_addr']
        cx_ram_load = cxram["sh_addr"]
        cx_ram_size = cxram["sh_size"]

    return sh_offset, sh_size, sh_load, cx_ram_size, cx_ram_load


def run_qemu(s1: socket.socket, s2: socket.socket, args: argparse.Namespace, use_bagl: bool) -> int:
    argv = ['qemu-arm-static']

    if args.debug:
        argv += ['-g', '1234', '-singlestep']

    argv += [launcher_path]

    if args.trace:
        argv += ['-t']

    argv += ['-m', args.model]

    if args.apiLevel:
        argv += ['-a', str(args.apiLevel)]
    else:
        argv += ['-k', str(args.sdk)]

    # load cxlib only if available for the specified api level or sdk
    if args.apiLevel:
        cxlib_filepath = f"cxlib/{args.model}-api-level-cx-{args.apiLevel}.elf"
    else:
        cxlib_filepath = f"cxlib/{args.model}-cx-{args.sdk}.elf"
    cxlib = str(resources.files(__package__) / cxlib_filepath)
    if os.path.exists(cxlib):
        sh_offset, sh_size, sh_load, cx_ram_size, cx_ram_load = get_cx_infos(cxlib)
        cxlib_args = f'{cxlib}:{sh_offset:#x}:{sh_size:#x}:{sh_load:#x}:{cx_ram_size:#x}:{cx_ram_load:#x}'
        argv += ['-c', cxlib_args]
    else:
        logger.warn(f"Cx lib {cxlib_filepath} not found")

    # for NBGL apps, fonts binary file is mandatory
    if not use_bagl:
        fonts_filepath = f"fonts/{args.model}-fonts-{args.apiLevel}.bin"
        fonts = str(resources.files(__package__) / fonts_filepath)
        if os.path.exists(fonts):
            argv += ['-f', fonts]
        else:
            logger.error(f"Fonts {fonts_filepath} not found")
            sys.exit(1)

    extra_ram = ''
    app_path = getattr(args, 'app.elf')
    for lib in [f'main:{app_path}'] + args.library:
        name, lib_path = lib.split(':')
        load_offset, load_size, stack, stack_size, ram_addr, ram_size, \
            text_load_addr, svc_call_address, svc_cx_call_address, \
            fonts_addr, fonts_size = get_elf_infos(lib_path, use_bagl)

        # Since binaries loaded as libs could also declare extra RAM page(s), collect them all
        if (ram_addr, ram_size) != (0, 0):
            arg = f'{ram_addr:#x}:{ram_size:#x}'
            if extra_ram and arg != extra_ram:
                logger.error("different extra RAM pages for main app and/or libraries!")
                sys.exit(1)
            extra_ram = arg
        lib_arg = f'{name}:{lib_path}:{load_offset:#x}:{load_size:#x}'
        lib_arg += f':{stack:#x}:{stack_size:#x}:{svc_call_address:#x}'
        lib_arg += f':{svc_cx_call_address:#x}:{text_load_addr:#x}'
        lib_arg += f':{fonts_addr:#x}:{fonts_size:#x}'
        argv.append(lib_arg)

    if args.model == 'blue':
        if args.rampage:
            extra_ram = args.rampage
        if extra_ram:
            argv.extend(['-r', extra_ram])

    pid = os.fork()
    if pid != 0:
        return pid

    # ensure qemu is killed when this Python script exits
    set_pdeath(signal.SIGTERM)

    s2.close()

    # replace stdin with the socket
    os.dup2(s1.fileno(), sys.stdin.fileno())

    # handle both BIP39 mnemonics and hex seeds
    if args.seed.startswith("hex:"):
        seed = bytes.fromhex(args.seed[4:])
    else:
        seed = mnemonic.Mnemonic.to_seed(args.seed)

    os.environ['SPECULOS_SEED'] = binascii.hexlify(seed).decode('ascii')

    if args.deterministic_rng:
        os.environ['RNG_SEED'] = args.deterministic_rng

    if args.user_private_key:
        os.environ['USER_PRIVATE_KEY'] = args.user_private_key
    if args.attestation_key:
        os.environ['ATTESTATION_PRIVATE_KEY'] = args.attestation_key

    logger.debug(f"executing qemu: {argv}")
    try:
        os.execvp(argv[0], argv)
    except FileNotFoundError:
        logger.error('failed to execute qemu: "%s" not found' % argv[0])
        sys.exit(1)
    sys.exit(0)


def setup_logging(args):
    logging.basicConfig(level=logging.INFO, format='%(asctime)s.%(msecs)03d:%(name)s: %(message)s', datefmt='%H:%M:%S')

    for arg in args.log_level:
        if ":" not in arg:
            logger.error(f"invalid --log argument {arg}")
            sys.exit(1)
        name, level = arg.split(":", 1)
        given_logger = logging.getLogger(name)
        try:
            given_logger.setLevel(level)
        except ValueError as e:
            logger.error(f"invalid --log argument: {e}")
            sys.exit(1)


def main(prog=None) -> int:

    parser = argparse.ArgumentParser(description='Emulate Ledger Nano/Blue apps.')
    parser.add_argument('app.elf', type=str, help='application path')
    parser.add_argument('--automation', type=str, help='Load a JSON document automating actions (prefix with "file:" '
                                                       'to specify a path')
    parser.add_argument('--color', default='MATTE_BLACK', choices=list(display.COLORS.keys()), help='Nano color')
    parser.add_argument('-d', '--debug', action='store_true', help='Wait gdb connection to port 1234')
    parser.add_argument('--deterministic-rng', default='', help='Seed the rng with a given value to produce '
                                                                'deterministic randomness')
    parser.add_argument('--user-private-key', default='',
                        help='32B in hex format, will be used as the user private keys')
    parser.add_argument('--attestation-key', default='', help='32B in hex format, will be used as the private '
                                                              'attestation key')
    parser.add_argument('-k', '--sdk', type=str, help='SDK version')
    parser.add_argument('-a', '--apiLevel', type=str, help='Api level')
    parser.add_argument('-l', '--library', default=[], action='append', help='Additional library (eg. '
                        'Bitcoin:app/btc.elf) which can be called through os_lib_call'
                        'You can also only pass the lib elf path if the lib name is included in the metadata')
    parser.add_argument('--log-level', default=[], action='append', help='Configure the logger levels (eg. usb:DEBUG), '
                                                                         'can be specified multiple times')
    parser.add_argument('-m', '--model', choices=list(display.MODELS.keys()))
    parser.add_argument('-r', '--rampage', help='Additional RAM page and size available to the app (eg. '
                                                '0x123000:0x100). Supersedes the internal probing for such page.')
    parser.add_argument('-s', '--seed', default=DEFAULT_SEED, help='BIP39 mnemonic or hex seed. Default to mnemonic: '
                                                                   'to use a hex seed, prefix it with "hex:"')
    parser.add_argument('-t', '--trace', action='store_true', help='Trace syscalls')
    parser.add_argument('-u', '--usb', default='hid', help='Configure the USB transport protocol, '
                                                           'either HID (default) or U2F')

    group = parser.add_argument_group('network arguments')
    group.add_argument('--apdu-port', default=9999, type=int, help='ApduServer TCP port')
    group.add_argument('--api-port', default=5000, type=int, help='Set the REST API TCP port (0 disables it)')
    group.add_argument('--automation-port', type=int, help='Forward text displayed on the screen to TCP clients')
    group.add_argument('--vnc-port', type=int, help='Start a VNC server on the specified port')
    group.add_argument('--vnc-password', type=str, help='VNC plain-text password (required for MacOS Screen Sharing)')
    group.add_argument('--button-port', type=int, help='Spawn a TCP server on the specified port to receive button '
                                                       'press (lLrR)')
    group.add_argument('--finger-port', type=int, help='Spawn a TCP server on the specified port to receive finger '
                                                       'touch (x,y,pressed)+')

    group = parser.add_argument_group('display arguments', 'These arguments might only apply to one of the display '
                                                           'method.')
    group.add_argument('--display', default='qt', choices=['headless', 'qt', 'text'])
    group.add_argument('--ontop', action='store_true', help='The window stays on top of all other windows')
    group.add_argument('--xy', action='store', help='Window position in "XxY" format (eg. "--xy 30x100").')
    group.add_argument('--keymap', action='store', help="Text UI keymap in the form of a string (e.g. 'was' => 'w' for "
                                                        "left button, 'a' right, 's' both). Default: arrow keys")
    group.add_argument('--progressive', action='store_true', help='Enable step-by-step rendering of graphical elements')
    group.add_argument('--zoom', help='Display pixel size.', type=int, choices=range(1, 11))

    if prog:
        parser.prog = prog
    args = parser.parse_args()

    if args.model:
        args.model = args.model.lower()

    # Initialize root logging level and handlers and module specific level if requested in command line
    setup_logging(args)

    # Init model and api_level if not specified from app elf metadata
    app_path = getattr(args, 'app.elf')
    binary = LedgerBinaryApp(app_path)
    if not args.model:
        if binary.sections.target is None:
            logger.error("Device model not detected from elf. Then it must be specified")
            sys.exit(1)
        else:
            args.model = "nanosp" if binary.sections.target == "nanos2" else binary.sections.target
            logger.warn(f"Device model detected from metadata: {args.model}")

    # 'bagl' is the default value of the binary.sections.sdk_graphics. We need to
    # manage the cases where it is NOT 'bagl' but the section does not exists yet
    if args.model in ["stax", "flex"]:
        use_bagl = False
    else:
        use_bagl = binary.sections.sdk_graphics == "bagl"

    if not args.apiLevel:
        if binary.sections.api_level is not None:
            args.apiLevel = binary.sections.api_level
            logger.warn(f"Api level detected from metadata: {args.apiLevel}")

    # Check args.apiLevel, 0 is an invalid value
    if args.apiLevel == "0":
        logger.error(f"Invalid api_level {args.apiLevel}")
        sys.exit(1)

    # Set SPECULOS_DETECTED_APPNAME env variable for proper emulation.
    # See sys_os_registry_get_current_app_tag() for corresponding usage.
    app_name = binary.sections.app_name
    app_version = binary.sections.app_version
    if app_name and app_version:
        os.environ["SPECULOS_DETECTED_APPNAME"] = f"{app_name}:{app_version}"

    # Retrieve lib app_name if available and check it against argument now optional
    libs = []
    for lib_arg in args.library:
        if ":" in lib_arg:
            lib_name, lib_path = lib_arg.split(":")
        else:
            lib_name = None
            lib_path = lib_arg

        elf_lib_name = LedgerBinaryApp(lib_path).sections.app_name

        if lib_name is None:
            if elf_lib_name is None:
                logger.error("Lib name not detected from elf. Then it must be specified")
                sys.exit(1)
            else:
                logger.warn(f"Lib name detected from metadata: {elf_lib_name}")
                lib_name = elf_lib_name
        else:
            if elf_lib_name is not None and elf_lib_name != lib_name:
                logger.error(f"Invalid lib name in {lib_path} ({elf_lib_name} vs {lib_name})")
                sys.exit(1)

        libs.append(f"{lib_name}:{lib_path}")
    args.library = libs

    # Check model and api_level against all lib elf metadata
    for path in [app_path] + [x.split(":")[1] for x in args.library]:
        binary = LedgerBinaryApp(path)

        elf_model = ("nanosp" if binary.sections.target == "nanos2" else binary.sections.target) or args.model
        if args.model != elf_model:
            logger.error(f"Invalid model in {path} ({elf_model} vs {args.model})")
            sys.exit(1)

        elf_api_level = binary.sections.api_level or "0"
        # Check args.apiLevel against elf api level. If elf api level == 0 (SDK master
        # reserved value) ignore it.
        if elf_api_level != "0" and args.apiLevel != elf_api_level:
            logger.error(f"Invalid api_level in {path} ({elf_api_level} vs {args.apiLevel})")
            sys.exit(1)

    rendering = seproxyhal.RENDER_METHOD.FLUSHED
    if args.progressive:
        rendering = seproxyhal.RENDER_METHOD.PROGRESSIVE

    if args.rampage:
        if args.model != 'blue':
            logger.error("extra RAM page arguments -r (--rampage) require '-m blue'")
            sys.exit(1)
        if not re.match('(0x)?[0-9a-fA-F]+:(0x)?[0-9a-fA-F]+$', args.rampage):
            logger.error("invalid ram page argument")
            sys.exit(1)

    if args.display == 'text' and args.model not in ['nanos', 'nanox', 'nanosp']:
        logger.error(f"unsupported model '{args.model}' with argument --display text")
        sys.exit(1)

    if args.ontop and args.display != 'qt':
        logger.error("--ontop can only be used with --display qt")
        sys.exit(1)

    if args.xy and args.display != 'qt':
        logger.error("--xy can only be used with --display qt")
        sys.exit(1)

    if args.zoom and args.display != 'qt':
        logger.error("--zoom can only be used with --display qt")
        sys.exit(1)

    if args.keymap and args.display != 'text':
        logger.error("--keymap can only be used with --display text")
        sys.exit(1)

    if args.vnc_password and not args.vnc_port:
        logger.error("--vnc-password can only be used with --vnc-port")
        sys.exit(1)

    ScreenNotifier: Type[display.DisplayNotifier]
    if args.display == 'text':
        from .mcu.screen_text import TextScreenNotifier as ScreenNotifier
    elif args.display == 'headless':
        from .mcu.headless import HeadlessNotifier as ScreenNotifier
    else:
        from .mcu.screen import QtScreenNotifier as ScreenNotifier

    if args.sdk and args.apiLevel:
        logger.error("Either SDK version or api level should be specified")
        sys.exit(1)

    if args.apiLevel is None and args.sdk is None:
        default_sdk = {
            "nanos": "2.1",
            "nanox": "2.0.2",
            "blue": "blue-2.2.5",
            "nanosp": "1.0.4"
        }
        args.sdk = default_sdk.get(args.model)

    if args.model == "nanosp" and args.sdk == "1.0.4":
        # NanoS+ 1.0.4 OS can be emulated as 1.0.3 OS
        args.sdk = "1.0.3"

    api_enabled = (args.api_port != 0)

    automation_path: Optional[automation.Automation] = None
    if args.automation:
        # TODO: remove this condition and all associated code in next major version
        logger.warn("--automation is deprecated, please use the REST API instead")
        automation_path = automation.Automation(args.automation)

    automation_server: Optional[BroadcastInterface] = None
    if args.automation_port:
        # TODO: remove this condition and all associated code in next major version
        logger.warn("--automation-port is deprecated, please use the REST API instead")
        if api_enabled:
            logger.warn("--automation-port is incompatible with the the API server, disabling the latter")
            api_enabled = False
        automation_server = AutomationServer(("0.0.0.0", args.automation_port), AutomationClient)
        automation_thread = threading.Thread(target=automation_server.serve_forever, daemon=True)
        automation_thread.start()

    if api_enabled:
        automation_server = EventsBroadcaster()

    s1, s2 = socket.socketpair()

    qemu_pid = run_qemu(s1, s2, args, use_bagl)
    s1.close()

    apdu = apdu_server.ApduServer(host="0.0.0.0", port=args.apdu_port)
    seph = seproxyhal.SeProxyHal(
        s2,
        model=args.model,
        use_bagl=use_bagl,
        automation=automation_path,
        automation_server=automation_server,
        transport=args.usb)

    button = None
    if args.button_port:
        logger.warn("--button-port is deprecated, please use the REST API instead")
        button = FakeButton(args.button_port)

    finger = None
    if args.finger_port:
        logger.warn("--finger-port is deprecated, please use the REST API instead")
        finger = FakeFinger(args.finger_port)

    vnc = None
    if args.vnc_port:
        screen_size = display.MODELS[args.model].screen_size
        vnc = VNC(args.vnc_port, screen_size, args.vnc_password)

    zoom = args.zoom
    if zoom is None:
        default_zoom = {
            "nanos": 2,
            "nanox": 2,
            "nanosp": 2,
            "blue": 1,
            "stax": 1,
            "flex": 1
        }
        zoom = default_zoom.get(args.model)

    x, y = None, None
    if args.xy:
        x, y = (int(i) for i in args.xy.split('x'))

    apirun: Optional[ApiRunner] = None
    if api_enabled:
        apirun = ApiRunner(args.api_port)

    display_args = DisplayArgs(args.color, args.model, args.ontop, rendering,
                               args.keymap, zoom, x, y, use_bagl)
    server_args = ServerArgs(apdu, apirun, button, finger, seph, vnc)
    screen_notifier = ScreenNotifier(display_args, server_args)

    if apirun is not None:
        assert automation_server is not None
        apirun.start_server_thread(screen_notifier, seph, automation_server)

    try:
        screen_notifier.run()
    except BaseException:
        # Will deal with exception triggered in the ScreenNotifier, including
        # KeyboardInterrupt (if not Qt display, else it will segfault)
        logger.exception("An error occurred")
        logger.critical("Stopping Speculos")
    finally:
        if apirun is not None:
            apirun.stop()

        s2.close()
        _, status = os.waitpid(qemu_pid, 0)
        qemu_exit_status = os.WEXITSTATUS(status)
        sys.exit(qemu_exit_status)



from __future__ import annotations

'''
Forward packets between an external application and the emulated device.

Internally, it makes use of a TCP socket server to allow these 2 components to
communicate.
'''

import errno
import logging
import socket
from typing import Optional

from .display import IODevice, DisplayNotifier


class ApduServer(IODevice):
    def __init__(self, host: str = '127.0.0.1', port: int = 9999, hid: bool = False):
        self.socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self.socket.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        self.socket.bind((host, port))  # lgtm [py/bind-socket-all-network-interfaces]
        self.socket.listen()
        self.client: Optional[ApduClient] = None

    @property
    def file(self):
        return self.socket

    def can_read(self, screen: DisplayNotifier):
        c, _ = self.file.accept()
        self.client = ApduClient(c)
        screen.add_notifier(self.client)

    def forward_to_client(self, packet: bytes):
        if self.client is not None:
            self.client.forward_to_client(packet)


class ApduClient(IODevice):
    def __init__(self, sock: socket.socket):
        self._socket = sock
        self.logger = logging.getLogger("apdu")

    @property
    def file(self):
        return self._socket

    def _recvall(self, size):
        data = b''
        while size > 0:
            try:
                tmp = self.file.recv(size)
            except ConnectionResetError:
                tmp = b''
            if len(tmp) == 0:
                self.logger.debug("connection with client closed")
                return None
            data += tmp
            size -= len(tmp)
        return data

    def recv_packet(self):
        data = self._recvall(4)
        if data is None:
            return None

        size = int.from_bytes(data, byteorder='big')
        packet = self._recvall(size)
        if packet is None:
            return None

        return packet

    def can_read(self, screen: DisplayNotifier) -> None:
        '''Forward APDU packet to the app'''
        packet = self.recv_packet()
        if packet is None:
            screen.remove_notifier(self.fileno)
            self.file.close()
            return

        self.logger.info("> {}".format(packet.hex()))
        screen.display.forward_to_app(packet)

    def forward_to_client(self, packet):
        '''Encode and forward APDU to the client.'''

        self.logger.info("< {}".format(packet.hex()))

        size = (len(packet) - 2) & 0xffffffff
        packet = size.to_bytes(4, 'big') + packet
        try:
            self.file.sendall(packet)
        except OSError as e:
            if e.errno == errno.EBADF:
                # the connection with the client was closed, ignore any error
                pass
            else:
                raise



import json
import logging
import requests
import socket
import sys
import subprocess
import time
from contextlib import contextmanager
from PIL import Image, ImageChops
from requests import Response
from typing import Generator, List, Optional, Tuple, Type
from types import TracebackType

logger = logging.getLogger("speculos-client")
logger.setLevel(logging.INFO)


class ApduException(Exception):
    def __init__(self, sw: int, data: bytes) -> None:
        super().__init__(f"Exception: invalid status 0x{sw:x}")
        self.sw = sw
        self.data = data


class ClientException(Exception):
    pass


def check_status_code(response: requests.Response, url: str) -> None:
    if response.status_code != 200:
        raise ClientException(
            f"HTTP request on {url} failed, status={response.status_code}, error={response.content!r}"
        )


class ApduResponse:
    def __init__(self, response: requests.Response) -> None:
        self.response = response

    def receive(self) -> bytes:
        check_status_code(self.response, "/apdu")
        data, status = split_apdu(bytes.fromhex(self.response.json()["data"]))
        if status != 0x9000:
            raise ApduException(status, data)
        return data


def split_apdu(data: bytes) -> Tuple[bytes, int]:
    if len(data) < 2:
        raise ClientException(f"APDU response length is shorter than 2 ({data!r})")
    status = int.from_bytes(data[-2:], "big")
    return data[:-2], status


def screenshot_equal(path1: str, path2: str, left: int = 0, upper: int = 0, right: int = 0, lower: int = 0) -> bool:
    """Compare two images and return True if they are equal."""

    with Image.open(path1) as img1:
        with Image.open(path2) as img2:
            if any([left, upper, right, lower]):
                img1_w, img1_h = img1.size
                img2_w, img2_h = img2.size
                img1 = img1.crop((left, upper, img1_w - right, img1_h - lower))
                img2 = img2.crop((left, upper, img2_w - right, img2_h - lower))
            diff_img = ImageChops.difference(img1, img2)
    return diff_img.getbbox() is None


class Api:
    def __init__(self, api_url: str) -> None:
        self.api_url = api_url
        self.timeout = 2000
        self.session = requests.Session()
        self.stream: Optional[Response] = None

    def open_stream(self) -> None:
        assert self.stream is None
        self.stream = self.session.get(f"{self.api_url}/events?stream=true", stream=True)
        check_status_code(self.stream, "/events")

    def close_stream(self) -> None:
        if self.stream:
            self.stream.close()
        self.stream = None

    def get_current_screen_content(self) -> dict:
        with self.session.get(f"{self.api_url}/events?currentscreenonly=true") as response:
            check_status_code(response, "/events")
            return response.json()

    def get_next_event(self) -> dict:
        """
        A subset of the event stream format is recognized by this function and the event is expected to be encoded in
        JSON.

        https://html.spec.whatwg.org/multipage/server-sent-events.html#parsing-an-event-stream
        """
        assert self.stream is not None
        data = b""
        while True:
            line = self.stream.raw.readline()
            if line == b"\n":
                break
            if not line.startswith(b"data: "):
                raise ClientException(f"unexpected stream event line ({line!r})")
            data += line[6:]

        event = json.loads(data)
        if not isinstance(event, dict):
            raise ClientException(f"Invalid event ({event})")

        return event

    def wait_for_text_event(self, text: str) -> dict:
        """Wait until an event containing the specified text is received."""

        while True:
            event = self.get_next_event()
            if text in event["text"]:
                break
        return event

    def press_and_release(self, button: str) -> None:
        assert button in ["left", "right", "both"]
        data = {"action": "press-and-release"}
        with self.session.post(f"{self.api_url}/button/{button}", json=data) as response:
            check_status_code(response, f"/button/{button}")

    def finger_touch(self,
                     x: int, y: int,
                     x2: Optional[int] = None, y2: Optional[int] = None,
                     delay: float = 0.5) -> None:
        data = {"action": "press-and-release", "x": x, "y": y, "delay": delay}
        if x2 is not None:
            data["x2"] = x2
        if y2 is not None:
            data["y2"] = y2
        with self.session.post(f"{self.api_url}/finger", json=data) as response:
            check_status_code(response, "/finger")

    def finger_swipe(self, x: int, y: int, direction: str, delay: float = 0.5) -> None:
        x2, y2 = x, y
        if direction == "up":
            y2 -= 10
        elif direction == "down":
            y2 += 10
        elif direction == "left":
            x2 -= 10
        elif direction == "right":
            x2 += 10
        data = {"action": "press-and-release", "x": x, "y": y, "x2": x2, "y2": y2, "delay": delay}
        with self.session.post(f"{self.api_url}/finger", json=data) as response:
            check_status_code(response, "/finger")

    def ticker_ctl(self, action: str) -> None:
        data = {"action": action}
        with self.session.post(f"{self.api_url}/ticker", json=data) as response:
            check_status_code(response, "/ticker")

    def get_screenshot(self) -> bytes:
        with self.session.get(f"{self.api_url}/screenshot") as response:
            check_status_code(response, "/screenshot")
            return response.content

    def _apdu_exchange(self, data: bytes, tick_timeout: int = 5 * 60 * 10) -> bytes:
        try:
            data_payload = {"data": data.hex(), "tick_timeout": tick_timeout}
            with self.session.post(f"{self.api_url}/apdu", json=data_payload) as response:
                apdu_response = ApduResponse(response)
                return apdu_response.receive()

        # TimeoutError exception in exchange function raises ChunkedEncodingError exception
        except requests.exceptions.ChunkedEncodingError:
            raise TimeoutError()

    def _apdu_exchange_nowait(self, data: bytes) -> requests.Response:
        return self.session.post(f"{self.api_url}/apdu", json={"data": data.hex()}, stream=True)

    def set_automation_rules(self, rules: dict) -> None:
        with self.session.post(f"{self.api_url}/automation", json=rules) as response:
            check_status_code(response, "/automation")


class SpeculosInstance:
    def __init__(self, app: str, args: Optional[List[str]] = None) -> None:
        self.app = app
        if args is None:
            self.args: List[str] = []
        else:
            self.args = list(args)
        self.process: Optional[subprocess.Popen] = None

        if "--display" not in self.args:
            self.args += ["--display", "headless"]

        if "--api-port" not in self.args:
            self.port = 5000
        else:
            n = self.args.index("--api-port")
            self.port = int(self.args[n + 1])

    def _wait_until_ready(self) -> None:
        connected = False
        for i in range(0, 50):
            try:
                s = socket.create_connection(("127.0.0.1", self.port))
                connected = True
                break
            except ConnectionRefusedError:
                time.sleep(0.1)

        if not connected:
            raise ClientException(f"Failed to connect to the speculos instance on port {self.port}")

        s.close()

    def start(self) -> None:
        cmd = [sys.executable or "python3", "-m", "speculos"] + self.args + [self.app]
        logger.info(f"starting speculos with command: {' '.join(cmd)}")
        self.process = subprocess.Popen(cmd)
        self._wait_until_ready()

    def stop(self) -> None:
        logger.info("stopping speculos")
        if self.process is None:
            return

        if self.process.poll() is None:
            self.process.terminate()
            try:
                self.process.wait(timeout=0.2)
            except subprocess.TimeoutExpired:
                self.process.kill()
                self.process.wait()

        self.process = None


class SpeculosClient(Api, SpeculosInstance):
    def __init__(self, app: str, args: List[str] = [], api_url: str = "http://127.0.0.1:5000") -> None:
        SpeculosInstance.__init__(self, app, args)
        Api.__init__(self, api_url)

    def start(self) -> None:
        """
        Start speculos to emulate the application. `stop` must be called to ensure proper
        termination of speculos.
        Note that context `with` statement can be used with `SpeculosClient` instance instead of
        using `start` and `stop` methods.
        """
        SpeculosInstance.start(self)
        self.open_stream()

    def stop(self) -> None:
        """ Terminate speculos instance started with `start` method. """
        self.close_stream()
        SpeculosInstance.stop(self)

    def __enter__(self) -> "SpeculosClient":
        self.start()
        return self

    def __exit__(
        self, exc_type: Optional[Type[BaseException]],
        exc_val: Optional[BaseException], exc_tb: Optional[TracebackType]
    ) -> None:
        self.stop()

    def apdu_exchange(
            self, cla: int, ins: int, data: bytes = b"", p1: int = 0, p2: int = 0,
            tick_timeout: int = 5 * 60 * 10) -> bytes:
        apdu = bytes([cla, ins, p1, p2, len(data)]) + data
        return Api._apdu_exchange(self, apdu, tick_timeout)

    @contextmanager
    def apdu_exchange_nowait(
        self, cla: int, ins: int, data: bytes = b"", p1: int = 0, p2: int = 0
    ) -> Generator[ApduResponse, None, None]:
        apdu = bytes([cla, ins, p1, p2, len(data)]) + data
        response = None
        try:
            response = Api._apdu_exchange_nowait(self, apdu)
            yield ApduResponse(response)
        finally:
            if response:
                response.close()



from dataclasses import dataclass
from typing import Any, Dict, NamedTuple, Optional, Tuple

Pixel = Tuple[int, int]


@dataclass
class Model:
    name: str
    screen_size: Tuple[int, int]
    box_position: Pixel
    box_size: Tuple[int, int]


MODELS: Dict[str, Model] = {
    'nanos': Model('Nano S', (128, 32), (20, 13), (100, 26)),
    'nanox': Model('Nano X', (128, 64), (5, 5), (10, 10)),
    'nanosp': Model('Nano SP', (128, 64), (5, 5), (10, 10)),
    'blue': Model('Blue', (320, 480), (13, 13), (26, 26)),
    'stax': Model('Stax', (400, 672), (13, 13), (26, 26)),
    'flex': Model('Flex', (480, 600), (13, 13), (26, 26)),
}


@dataclass
class DisplayArgs:
    color: str
    model: str
    ontop: bool
    rendering: bool
    keymap: str
    pixel_size: int
    x: Optional[int]
    y: Optional[int]
    use_bagl: bool


class ServerArgs(NamedTuple):
    apdu: Any  # ApduServer
    apirun: Any  # Optional[ApiRunner]
    button: Any  # Optional[FakeButton]
    finger: Any  # Optional[FakeFinger]
    seph: Any  # SeProxyHal
    vnc: Any  # Optional[VNC]



import pytest
import socket
import time


class Vnc:
    def __init__(self, port):
        self.s = self.connect(port)

    def connect(self, port):
        for i in range(0, 5):
            try:
                s = socket.create_connection(("127.0.0.1", port), timeout=0.5)
                connected = True
                break
            except ConnectionRefusedError:
                time.sleep(0.2)
                connected = False

        assert connected
        return s

    def auth(self, password=None):
        # recv server protocol version
        data = self.s.recv(4096)
        # send client protocol version
        self.s.sendall(data)
        # receive security types supported
        data = self.s.recv(2)
        if not password:
            assert data == b"\x01\x01"
        else:
            assert data == b"\x01\x02"


@pytest.fixture(scope="function")
def client(client_vnc):
    return client_vnc


@pytest.mark.additional_args("--vnc-port", "5900")
def test_vnc_no_password(client):
    vnc = Vnc(5900)
    vnc.auth()


@pytest.mark.additional_args("--vnc-port", "5900", "--vnc-password", "secret")
def test_vnc_with_password(client):
    vnc = Vnc(5900)
    vnc.auth("secret")



import json
import logging
import threading
from typing import Dict, Generator, List, Optional, Tuple, Union
from dataclasses import asdict
from flask import stream_with_context, Response
from flask_restful import inputs, reqparse

from speculos.observer import BroadcastInterface, ObserverInterface, TextEvent
from .restful import AppResource


class EventsBroadcaster(BroadcastInterface):
    """This used to be the 'Automation Server'."""

    def __init__(self) -> None:
        super().__init__()
        self.screen_content: List[TextEvent] = []
        self.events: List[TextEvent] = []
        self.condition = threading.Condition()
        self.logger = logging.getLogger("events")

    def clear_events(self) -> None:
        self.logger.debug("Clearing events")
        self.screen_content = []

    def broadcast(self, event: TextEvent) -> None:
        if event.clear:
            self.clear_events()
            return

        self.logger.debug("events: broadcasting %s to %d client(s)", asdict(event), len(self.clients))
        self.screen_content.append(event)
        self.events.append(event)
        for client in self.clients:
            client.send_screen_event(event)
        with self.condition:
            self.condition.notify_all()


class EventClient(ObserverInterface):
    def __init__(self, broadcaster: EventsBroadcaster) -> None:
        self.events: List[TextEvent] = []
        self._broadcaster = broadcaster

    def generate(self) -> Generator[bytes, None, None]:
        try:
            # force headers to be sent
            yield b""

            while True:
                with self._broadcaster.condition:
                    self._broadcaster.condition.wait(1)

                while self.events:
                    event = self.events.pop(0)
                    data = json.dumps(asdict(event))
                    # Format the event as specified in the specification:
                    # https://html.spec.whatwg.org/multipage/server-sent-events.html#parsing-an-event-stream
                    yield f"data: {data}\n\n".encode()
        finally:
            self._broadcaster.remove_client(self)

    def send_screen_event(self, event: TextEvent) -> None:
        self.events.append(event)


class Events(AppResource):
    def __init__(self, *args, automation_server: Optional[EventsBroadcaster] = None, **kwargs) -> None:
        if automation_server is None:
            raise RuntimeError("Argument 'automation_server' must not be None")
        self._broadcaster = automation_server
        self.parser = reqparse.RequestParser()
        self.parser.add_argument("stream", type=inputs.boolean, default=False, location='values')
        self.parser.add_argument("currentscreenonly", type=inputs.boolean, default=False, location='values')
        super().__init__(*args, **kwargs)

    def get(self) -> Union[Response, Tuple[Dict[str, List], int]]:
        args = self.parser.parse_args()
        if args.stream:
            client = EventClient(self._broadcaster)
            self._broadcaster.add_client(client)
            return Response(stream_with_context(client.generate()), content_type="text/event-stream")
        elif args.currentscreenonly:
            event_list = self._broadcaster.screen_content
        else:
            event_list = self._broadcaster.events
        return {"events": [asdict(e) for e in event_list]}, 200

    def delete(self) -> Tuple[Dict, int]:
        self._broadcaster.events.clear()
        return {}, 200



import json
import threading
import jsonschema
from flask import stream_with_context, Response, request
from typing import Generator, Optional

from speculos.resources_importer import get_resource_schema_as_json
from ..mcu.seproxyhal import SeProxyHal
from .restful import SephResource


class APDUBridge:
    def __init__(self, seph: SeProxyHal):
        # We want to be notified when APDU response is transmitted from the SE
        self.endpoint_lock = threading.Lock()
        self.response_condition = threading.Condition()
        self._seph = seph
        self._seph.apdu_callbacks.append(self.seph_apdu_callback)
        self.response: Optional[bytes]

    def exchange(self, data: bytes, tick_timeout: int = 5 * 60 * 10) -> Generator[bytes, None, None]:
        # force headers to be sent
        yield b""

        tick_count_before_exchange = self._seph.get_tick_count()

        with self.endpoint_lock:  # Lock for a command/response for one client
            with self.response_condition:
                self.response = None
            self._seph.to_app(data)
            with self.response_condition:
                while self.response is None:
                    self.response_condition.wait(0.1)
                    exchange_tick_count = self._seph.get_tick_count() - tick_count_before_exchange

                    if tick_timeout != 0 and exchange_tick_count > tick_timeout:
                        raise TimeoutError()
            yield json.dumps({"data": self.response.hex()}).encode()

    def seph_apdu_callback(self, data: bytes) -> None:
        """
        Called by seph when data is transmitted by the SE. That data should
        be the response to a prior APDU request
        """
        with self.response_condition:
            self.response = data
            self.response_condition.notify()


class APDU(SephResource):
    schema = get_resource_schema_as_json("api", "apdu.schema")

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._bridge = APDUBridge(self.seph)

    def post(self):
        args = request.get_json(force=True)
        try:
            jsonschema.validate(instance=args, schema=self.schema)
        except jsonschema.exceptions.ValidationError as e:
            return {"error": f"{e}"}, 400

        data = bytes.fromhex(args.get("data"))

        if "tick_timeout" in args:
            tick_timeout = args["tick_timeout"]
            return Response(stream_with_context(self._bridge.exchange(data, tick_timeout)),
                            content_type="application/json")

        return Response(stream_with_context(self._bridge.exchange(data)), content_type="application/json")



import logging
import sys
import threading
import time
from collections import namedtuple
from enum import IntEnum
from socket import socket
from typing import Callable, List, Optional, Tuple

from speculos.observer import BroadcastInterface, TextEvent
from . import usb
from .automation import Automation
from .display import DisplayNotifier, IODevice
from .nbgl import NBGL
from .ocr import OCR
from .readerror import ReadError


class SephTag(IntEnum):
    BUTTON_PUSH_EVENT = 0x05
    FINGER_EVENT = 0x0c
    DISPLAY_PROCESSED_EVENT = 0x0d
    TICKER_EVENT = 0x0e
    CAPDU_EVENT = 0x16

    MCU = 0x31
    TAG_BLE_SEND = 0x38
    TAG_BLE_RADIO_POWER = 0x44
    SE_POWER_OFF = 0x46
    USB_CONFIG = 0x4f
    USB_EP_PREPARE = 0x50

    REQUEST_STATUS = 0x52
    RAPDU = 0x53
    PLAY_TUNE = 0x56

    DBG_SCREEN_DISPLAY_STATUS = 0x5e
    PRINTC_STATUS = 0x5f

    GENERAL_STATUS = 0x60
    GENERAL_STATUS_LAST_COMMAND = 0x0000
    SCREEN_DISPLAY_STATUS = 0x65
    PRINTF_STATUS = 0x66
    SCREEN_DISPLAY_RAW_STATUS = 0x69

    FINGER_EVENT_TOUCH = 0x01
    FINGER_EVENT_RELEASE = 0x02

    # Speculos only, defined in speculos/src/bolos/bagl.c
    BAGL_DRAW_RECT = 0xF1
    BAGL_DRAW_BITMAP = 0xF2

    # Speculos only, defined in speculos/src/bolos/nbgl.c
    NBGL_DRAW_RECT = 0xFA
    NBGL_REFRESH = 0xFB
    NBGL_DRAW_LINE = 0xFC
    NBGL_DRAW_IMAGE = 0xFD
    NBGL_DRAW_IMAGE_FILE = 0xFE
    NBGL_DRAW_IMAGE_RLE = 0xFF


TICKER_DELAY = 0.1

RenderMethods = namedtuple('RenderMethods', 'PROGRESSIVE FLUSHED')
RENDER_METHOD = RenderMethods(0, 1)


class TimeTickerDaemon(threading.Thread):
    def __init__(self, add_tick: Callable, wait_until_tick_is_processed: Callable, *args, **kwargs):
        """
        Initializes the Ticker Daemon
        The goal of this daemon is to emulate the time spent in the application by sending it
        ticker events at a regular interval TICKER_DELAY.
        This daemon can be paused and resumed through its API to stop the flow of time in the MCU

        :param add_tick: The callback function to add a Ticker Event to the packet manager
        :type add_tick: Backend
        """
        super().__init__(name="time_ticker", daemon=True)
        self.paused = False
        self._paused = False
        self._resume_cond = threading.Condition()
        self.add_tick = add_tick
        self.wait_until_tick_is_processed = wait_until_tick_is_processed

    def pause(self):
        """
        Pause time emulation done by the daemon, no ticker event will be sent until resume
        """
        self.paused = True

        # Wait until the daemon is really paused before returning.
        # To make sure last daemon tick has been sent and fully processed.
        while not self._paused:
            time.sleep(0.1)

    def resume(self):
        """
        Resume time emulation done by the daemon
        """
        self.paused = False
        with self._resume_cond:
            self._resume_cond.notify()

    def _wait_if_time_paused(self):
        """
        Internal function to handle the pause
        """
        while self.paused:
            self.wait_until_tick_is_processed()
            self._paused = True
            with self._resume_cond:
                self._resume_cond.wait()
        self._paused = False

    def run(self):
        """
        Main thread function
        """
        while True:
            self._wait_if_time_paused()
            self.add_tick()
            time.sleep(TICKER_DELAY)


class SocketHelper(threading.Thread):
    def __init__(self, sock: socket, status_event: threading.Event, *args, **kwargs):
        super().__init__(name="packet", daemon=True)
        self.socket = sock
        self.sending_lock = threading.Lock()
        self.queue_condition = threading.Condition()
        self.queue: List[Tuple[SephTag, bytes]] = []
        self.status_event = status_event
        self.logger = logging.getLogger("seproxyhal.packet")
        self.stop = False
        self.ticks_count = 0
        self.tick_requested = False

    def _recvall(self, size: int):
        data = b''
        while size > 0:
            try:
                tmp = self.socket.recv(size)
            except ConnectionResetError:
                tmp = b''

            if len(tmp) == 0:
                self.logger.debug("fd closed")
                return None
            data += tmp
            size -= len(tmp)
        return data

    def get_tick_count(self):
        return self.ticks_count

    def read_packet(self):
        data = self._recvall(3)
        if data is None:
            return None, None, None

        tag = data[0]
        size = int.from_bytes(data[1:3], 'big')

        data = self._recvall(size)
        if data is None:
            return None, None, None
        assert len(data) == size

        return tag, size, data

    def send_packet(self, tag: SephTag, data: bytes = b''):
        """Send a packet to the app."""

        size: bytes = len(data).to_bytes(2, 'big')
        packet: bytes = tag.to_bytes(1, 'big') + size + data
        with self.sending_lock:
            self.logger.debug("send {}" .format(packet.hex()))
            try:
                self.socket.sendall(packet)
            except BrokenPipeError:
                self.stop = True
            except OSError:
                self.stop = True

    def queue_packet(self, tag: SephTag, data: bytes = b'', priority: bool = False):
        """
        Append a packet to the queue of packets to be sent.

        This function is meant to called by other threads.
        """

        if not priority:
            self.queue.append((tag, data))
        else:
            # Some status packets expect a specific event to be answered before
            # any other events. Put it at the beginning of the queue.
            self.queue.insert(0, (tag, data))

        # notify this thread that a new packet is available
        with self.queue_condition:
            self.queue_condition.notify()

    def wait_until_tick_is_processed(self):
        # Wait until the app has finished processing the tick
        while self.tick_requested or self.queue or not self.status_event.is_set():
            self.status_event.wait()

    def add_tick(self, wait_until_tick_is_processed=False):
        """Request sending of a ticker event to the app"""
        self.tick_requested = True

        # notify this thread that a new event is available
        with self.queue_condition:
            self.queue_condition.notify()

        if wait_until_tick_is_processed:
            self.wait_until_tick_is_processed()

    def run(self):
        while not self.stop:

            # wait for a event in the queue or a tick to be available
            with self.queue_condition:
                while len(self.queue) == 0 and not self.tick_requested:
                    self.queue_condition.wait()

            # wait for a status notification
            while not self.status_event.is_set():
                self.status_event.wait()
            self.status_event.clear()

            if len(self.queue):
                tag, data = self.queue.pop(0)
            elif self.tick_requested:
                tag, data = SephTag.TICKER_EVENT, b''
            else:
                raise RuntimeError("Unexpected state: no ticker nor event to send on socket")

            self.send_packet(tag, data)

            if tag == SephTag.TICKER_EVENT:
                self.ticks_count += 1
                self.tick_requested = False

        self.logger.debug("exiting")


class SeProxyHal(IODevice):
    def __init__(self,
                 sock: socket,
                 model: str,
                 use_bagl: bool,
                 automation: Optional[Automation] = None,
                 automation_server: Optional[BroadcastInterface] = None,
                 transport: str = 'hid'):
        self._socket = sock
        self.logger = logging.getLogger("seproxyhal")
        self.printf_queue = ''
        self.automation = automation
        self.automation_server = automation_server
        self.events: List[TextEvent] = []
        self.refreshed = False

        self.status_event = threading.Event()
        self.socket_helper = SocketHelper(self._socket, self.status_event)
        self.socket_helper.start()

        self.time_ticker_thread = TimeTickerDaemon(self.socket_helper.add_tick,
                                                   self.socket_helper.wait_until_tick_is_processed)
        self.time_ticker_thread.start()

        self.usb = usb.USB(self.socket_helper.queue_packet, transport=transport)

        self.ocr = OCR(model, use_bagl)

        self.use_bagl = use_bagl

        # A list of callback methods when an APDU response is received
        self.apdu_callbacks: List[Callable[[bytes], None]] = []

    @property
    def file(self):
        return self._socket

    def apply_automation_helper(self, event: TextEvent):
        if self.automation_server:
            self.automation_server.broadcast(event)

        if self.automation:
            actions = self.automation.get_actions(event.text, event.x, event.y)
            for action in actions:
                self.logger.debug(f"applying automation {action}")
                key, args = action[0], action[1:]
                if key == "button":
                    self.handle_button(*args)
                elif key == "finger":
                    self.handle_finger(*args)
                elif key == "setbool":
                    self.automation.set_bool(*args)
                elif key == "exit":
                    self.file.close()
                    sys.exit(0)
                else:
                    assert False

    def apply_automation(self):
        for event in self.events:
            self.apply_automation_helper(event)
        self.events = []

    def _cleanup(self, notifier: DisplayNotifier):
        notifier.remove_notifier(self.fileno)
        self.file.close()

    def can_read(self, screen: DisplayNotifier):
        '''
        Handle packet sent by the app.

        This function is called thanks to a screen QSocketNotifier.
        '''
        tag, size, data = self.socket_helper.read_packet()
        if data is None:
            self._cleanup(screen)
            raise ReadError("fd closed")

        self.logger.debug(f"received (tag: {tag:#04x}, size: {size:#04x}): {data!r}")

        if tag == SephTag.GENERAL_STATUS:
            if int.from_bytes(data[:2], 'big') == SephTag.GENERAL_STATUS_LAST_COMMAND:
                if self.refreshed:
                    self.refreshed = False

                    # Update the screenshot, we'll upload its associated events shortly
                    screen.display.gl.update_screenshot()
                    screen.display.gl.update_public_screenshot()

                if self.use_bagl and screen.display.screen_update():
                    if screen.display.model in ["nanox", "nanosp"]:
                        self.events += self.ocr.get_events()
                elif not self.use_bagl:
                    self.events += self.ocr.get_events()

                # Apply automation rules after having received a GENERAL_STATUS_LAST_COMMAND tag. It allows the
                # screen to be updated before broadcasting the events.
                if self.events:
                    self.apply_automation()

                # signal the sending thread that a status has been received
                self.status_event.set()

            else:
                self.logger.error(f"unknown subtag: {data[:2]!r}")
                sys.exit(0)

        elif tag in [SephTag.SCREEN_DISPLAY_STATUS,
                     SephTag.DBG_SCREEN_DISPLAY_STATUS,
                     SephTag.BAGL_DRAW_RECT]:
            self.logger.debug(f"DISPLAY_STATUS {data!r}")
            if screen.display.model not in ["nanox", "nanosp"] or tag == SephTag.BAGL_DRAW_RECT:
                events = screen.display.display_status(data)
                self.events += events
            if tag != SephTag.BAGL_DRAW_RECT:
                self.socket_helper.send_packet(SephTag.DISPLAY_PROCESSED_EVENT)

        elif tag in [SephTag.SCREEN_DISPLAY_RAW_STATUS, SephTag.BAGL_DRAW_BITMAP]:
            self.logger.debug("SephTag.SCREEN_DISPLAY_RAW_STATUS")
            screen.display.display_raw_status(data)
            if screen.display.model in ["nanox", "nanosp"]:
                self.ocr.analyze_bitmap(data)
            if tag != SephTag.BAGL_DRAW_BITMAP:
                self.socket_helper.send_packet(SephTag.DISPLAY_PROCESSED_EVENT)
            if screen.display.rendering == RENDER_METHOD.PROGRESSIVE:
                screen.display.screen_update()

        elif tag == SephTag.PRINTF_STATUS or tag == SephTag.PRINTC_STATUS:
            for b in [chr(b) for b in data]:
                if b == '\n':
                    self.logger.info(f"printf: {self.printf_queue}")
                    self.printf_queue = ''
                else:
                    self.printf_queue += b
            if screen.display.model in ["blue"]:
                self.socket_helper.send_packet(SephTag.DISPLAY_PROCESSED_EVENT)

        elif tag == SephTag.RAPDU:
            screen.display.forward_to_apdu_client(data)
            for c in self.apdu_callbacks:
                c(data)

        elif tag == SephTag.USB_CONFIG:
            self.usb.config(data)

        elif tag == SephTag.USB_EP_PREPARE:
            data = self.usb.prepare(data)
            if data:
                for c in self.apdu_callbacks:
                    c(data)
                screen.display.forward_to_apdu_client(data)

        elif tag == SephTag.MCU:
            pass

        elif tag == SephTag.TAG_BLE_SEND:
            pass

        elif tag == SephTag.TAG_BLE_RADIO_POWER:
            self.logger.warn("ignoring BLE tag")

        elif tag == SephTag.SE_POWER_OFF:
            self.logger.warn("received tag SE_POWER_OFF, exiting")
            self._cleanup(screen)
            raise ReadError("SE_POWER_OFF")

        elif tag == SephTag.REQUEST_STATUS:
            # Ignore calls to io_seproxyhal_request_mcu_status()
            pass

        elif tag == SephTag.PLAY_TUNE:
            pass

        elif tag == SephTag.NBGL_DRAW_RECT:
            assert isinstance(screen.display.gl, NBGL)
            self.events += screen.display.gl.hal_draw_rect(data)

        elif tag == SephTag.NBGL_REFRESH:
            assert isinstance(screen.display.gl, NBGL)
            screen.display.gl.refresh(data)
            # Stax/Flex only
            # We have refreshed the screen, remember it for the next time we have SephTag.GENERAL_STATUS
            # then we'll perform a screen update and make public the resulting screenshot
            self.refreshed = True

        elif tag == SephTag.NBGL_DRAW_LINE:
            assert isinstance(screen.display.gl, NBGL)
            screen.display.gl.hal_draw_line(data)

        elif tag == SephTag.NBGL_DRAW_IMAGE:
            assert isinstance(screen.display.gl, NBGL)
            self.ocr.analyze_bitmap(data)
            screen.display.gl.hal_draw_image(data)

        elif tag == SephTag.NBGL_DRAW_IMAGE_RLE:
            assert isinstance(screen.display.gl, NBGL)
            self.ocr.analyze_bitmap(data)
            screen.display.gl.hal_draw_image_rle(data)

        elif tag == SephTag.NBGL_DRAW_IMAGE_FILE:
            assert isinstance(screen.display.gl, NBGL)
            screen.display.gl.hal_draw_image_file(data)

        else:
            self.logger.error(f"unknown tag: {tag:#x}")
            sys.exit(0)

    def handle_button(self, button: int, pressed: bool):
        '''Forward button press/release from the GUI to the app.'''

        if pressed:
            self.socket_helper.queue_packet(SephTag.BUTTON_PUSH_EVENT, (button << 1).to_bytes(1, 'big'))
        else:
            self.socket_helper.queue_packet(SephTag.BUTTON_PUSH_EVENT, (0 << 1).to_bytes(1, 'big'))

    def handle_finger(self, x: int, y: int, pressed: bool):
        '''Forward finger press/release from the GUI to the app.'''

        if pressed:
            packet = SephTag.FINGER_EVENT_TOUCH.to_bytes(1, 'big')
        else:
            packet = SephTag.FINGER_EVENT_RELEASE.to_bytes(1, 'big')
        packet += x.to_bytes(2, 'big')
        packet += y.to_bytes(2, 'big')

        self.socket_helper.queue_packet(SephTag.FINGER_EVENT, packet)

    def handle_ticker_request(self, action):
        if action == "pause":
            self.time_ticker_thread.pause()
        elif action == "resume":
            self.time_ticker_thread.resume()
        elif action == "single-step":
            self.time_ticker_thread.add_tick(wait_until_tick_is_processed=True)

    def handle_wait(self, delay: float):
        '''Wait for a specified delay, taking account real time seen by the app.'''
        expected_ticks = int(delay / TICKER_DELAY)
        if not self.time_ticker_thread.paused:
            start = self.socket_helper.ticks_count
            while (self.socket_helper.ticks_count - start) < expected_ticks:
                time.sleep(TICKER_DELAY)
        else:
            for _ in range(expected_ticks):
                self.time_ticker_thread.add_tick(wait_until_tick_is_processed=True)

    def to_app(self, packet: bytes):
        '''
        Forward raw APDU to the app.

        Packets can be forwarded directly to the SE thanks to
        SephTag.CAPDU_EVENT, but it doesn't work with messages are larger
        than G_io_seproxyhal_spi_buffer. Emulating a basic USB stack is more
        reliable, see issue #11.
        '''

        if packet.startswith(b'RAW!') and len(packet) > 4:
            tag, packet = packet[4], packet[5:]
            self.socket_helper.queue_packet(SephTag(tag), packet)
        else:
            self.usb.xfer(packet)

    def get_tick_count(self):
        return self.socket_helper.get_tick_count()



import json
from pathlib import Path
from platform import python_version_tuple
from typing import Dict

major, minor, _ = python_version_tuple()
assert major == "3"

if int(minor) <= 8:
    import importlib_resources as resources
else:
    import importlib.resources as resources


def get_resources_path(module: str, filename: str) -> Path:
    return resources.files(__package__) / module / "resources" / filename


def get_resource_schema_as_json(module: str, filename: str) -> Dict:
    with get_resources_path(module, filename).open("rb") as fp:
        schema = json.load(fp)
    return schema



#!/usr/bin/env python3
"""Install Speculos"""
import pathlib
import tempfile
from distutils.spawn import find_executable
from setuptools.command.build_py import build_py as _build_py
from setuptools import setup


class BuildSpeculos(_build_py):
    """
    Extend "setup.py build_py" to build Speculos launcher and VNC server using cmake.

    This command requires some system dependencies (ARM compiler, libvncserver
    headers...) which are documented on https://speculos.ledger.com/installation/build.html

    distutils documentation about extending the build command:
    https://docs.python.org/3.8/distutils/extending.html#integrating-new-commands
    """

    def run(self):
        if not find_executable("cmake"):
            raise RuntimeError("cmake is not found and is required to build Speculos")

        if not self.dry_run:
            pathlib.Path(self.build_lib).mkdir(parents=True, exist_ok=True)
            with tempfile.TemporaryDirectory(prefix="build-", dir=self.build_lib) as build_dir:
                self.spawn(
                    [
                        "cmake",
                        "-H.",
                        "-B" + build_dir,
                        "-DCMAKE_BUILD_TYPE=Release",
                        "-DBUILD_TESTING=0",
                        "-DWITH_VNC=1",
                    ]
                )
                self.spawn(["cmake", "--build", build_dir])

        super().run()


setup(
    cmdclass={
        "build_py": BuildSpeculos,
    },
)



"""Tests for Vault app."""

import pytest


@pytest.fixture(scope="module")
def client(client_ram_page):
    return client_ram_page


def test_access_ram_page(client):
    client.apdu_exchange(0xe0, 0x01, b"")
